@@ -1,9 +1,9 @@
 import numpy as np
 import os
+import logging
 
 from kgcnn.data.base import MemoryGraphDataset
 
-
 # TUDataset: A collection of benchmark datasets for learning with graphs
 # by Christopher Morris and Nils M. Kriege and Franka Bause and Kristian Kersting and Petra Mutzel and Marion Neumann
 # http://graphlearning.io
@@ -12,63 +12,49 @@ from kgcnn.data.base import MemoryGraphDataset
 class GraphTUDataset(MemoryGraphDataset):
     r"""Base class for loading graph datasets published by `TU Dortmund University
     <https://chrsmrrs.github.io/datasets>`_. Datasets contain non-isomorphic graphs. This general base class has
-    functionality to load TUDatasets in a generic way.
+    functionality to load TUDatasets in a generic way. The datasets are already in a graph-like format and do not
+    need further processing via e.g. `prepare data`.
 
     .. note::
-        Note that sub-classes of `GraphTUDataset2020` in :obj:``kgcnn.data.datasets`` downloads datasets,
-        There are also further TU-datasets in :obj:``kgcnn.data.datasets``, if further processing is used in literature.
+        Note that sub-classes of `GraphTUDataset2020` in :obj:`kgcnn.data.datasets` downloads datasets,
+        There are also further TU-datasets in :obj:`kgcnn.data.datasets`, if further processing is used in literature.
         Not all datasets can provide all types of graph properties like `edge_attributes` etc.
 
     """
 
     def __init__(self, data_directory: str = None, dataset_name: str = None, file_name: str = None,
-                 verbose: int = 1):
-        r"""Initialize a `GraphTUDataset` instance from file.
+                 file_directory: str = None, verbose: int = 10):
+        r"""Initialize a :obj:`GraphTUDataset` instance from file.
 
         Args:
+            data_directory (str): Full path to directory of the dataset. Default is None.
             file_name (str): Filename for reading into memory. Not used for general TUDataset, since there are multiple
                 files with a prefix and pre-defined suffix. Default is None.
-            data_directory (str): Full path to directory containing all txt-files. Default is None.
+            file_directory (str): Name or relative path from :obj:`data_directory` to a directory containing sorted
+                files. Default is None.
             dataset_name (str): Name of the dataset. Important for base-name for naming of files. Default is None.
-            verbose (int): Print progress or info for processing, where 0 is silent. Default is 1.
+            verbose (int): Logging level. Default is 10.
         """
-        self.file_name = file_name
-        self.data_directory = data_directory
-        self.dataset_name = dataset_name
+        MemoryGraphDataset.__init__(self, data_directory=data_directory, dataset_name=dataset_name,
+                                    file_name=file_name, verbose=verbose, file_directory=file_directory)
 
-        MemoryGraphDataset.__init__(self, verbose=verbose)
-
-    def read_in_memory(self, file_name: str = None, data_directory: str = None, dataset_name: str = None,
-                       verbose: int = 1):
+    def read_in_memory(self):
         r"""Read the TUDataset into memory. The TUDataset is stored in disjoint representations. The data is cast
         to a list of separate graph properties for `MemoryGraphDataset`.
 
-        Args:
-            file_name (str): Filename for reading into memory. Not used for general TUDataset, since there are multiple
-                files with a prefix and pre-defined suffix. Default is None.
-            data_directory (str): Full path to directory containing all txt-files. Default is None.
-            dataset_name (str): Name of the dataset. Important for base-name naming of files. Default is None.
-            verbose (int): Print progress or info for processing, where 0 is silent. Default is 1.
-
         Returns:
             self
         """
-        if file_name is not None:
-            self.file_name = file_name
-        if data_directory is not None:
-            self.data_directory = data_directory
-        if self.dataset_name is None:
-            self.dataset_name = dataset_name
-
         if self.dataset_name is not None and self.data_directory is not None:
             path = os.path.realpath(self.data_directory)
             name_dataset = self.dataset_name
+            if self.file_directory is not None:
+                path = os.path.join(path, self.file_directory)
         else:
-            print("ERROR:kgcnn: Dataset needs name {0} and path {1}.".format(self.dataset_name, self.data_directory))
-            return None
+            self.error("Dataset needs name {0} and path {1}.".format(self.dataset_name, self.data_directory))
+            return self
 
-        if verbose > 1:
-            print("INFO:kgcnn: Reading dataset to memory with name %s" % str(self.dataset_name))
+        self.info("Reading dataset to memory with name %s" % str(self.dataset_name))
 
         # Define a graph with indices
         # They must be defined
@@ -114,13 +100,12 @@ class GraphTUDataset(MemoryGraphDataset):
         num_graphs = np.amax(g_n_id)
         if g_labels is not None:
             if len(g_labels) != num_graphs:
-                print("ERROR:kgcnn: Wrong number of graphs, not matching graph labels, {0}, {1}".format(len(g_labels),
-                                                                                                        num_graphs))
+                self.error(
+                    "Wrong number of graphs, not matching graph labels, {0}, {1}".format(len(g_labels), num_graphs))
 
         # shift index, should start at 0 for python indexing
         if int(np.amin(g_n_id)) == 1 and int(np.amin(g_a)) == 1:
-            if verbose > 0:
-                print("INFO:kgcnn: Shift start of graph id to zero for %s to match python indexing." % name_dataset)
+            self.info("Shift start of graph id to zero for %s to match python indexing." % name_dataset)
             g_a = g_a - 1
             g_n_id = g_n_id - 1
 
@@ -161,23 +146,42 @@ class GraphTUDataset(MemoryGraphDataset):
             all_cons.append(np.sum(is_cons == False))
         all_cons = np.array(all_cons)
 
-        if verbose > 0:
-            print("INFO:kgcnn: Graph index which has unconnected", np.arange(len(all_cons))[all_cons > 0], "with",
-                  all_cons[all_cons > 0], "in total", len(all_cons[all_cons > 0]))
+        self.info("Graph index which has unconnected %s with %s in total %s" % (
+            np.arange(len(all_cons))[all_cons > 0], all_cons[all_cons > 0], len(all_cons[all_cons > 0])))
 
         node_degree = [np.zeros(x, dtype="int") for x in graphlen]
         for i, x in enumerate(edge_indices):
             nod_id, nod_counts = np.unique(x[:, 0], return_counts=True)
             node_degree[i][nod_id] = nod_counts
 
-        self.node_degree = node_degree
-        self.node_attributes = n_attr
-        self.edge_attributes = e_attr
-        self.graph_attributes = g_attr
-        self.edge_indices = edge_indices
-        self.node_labels = n_labels
-        self.edge_labels = e_labels
-        self.graph_labels = g_labels
-        self.length = num_graphs
+        self.assign_property("node_degree", node_degree)
+        self.assign_property("node_attributes", n_attr)
+        self.assign_property("edge_attributes", e_attr)
+        self.assign_property("graph_attributes", g_attr)
+        self.assign_property("edge_indices", edge_indices)
+        self.assign_property("node_labels", n_labels)
+        self.assign_property("edge_labels", e_labels)
+        self.assign_property("graph_labels", [x for x in g_labels])
 
         return self
+
+    @staticmethod
+    def read_csv_simple(filepath: str, delimiter: str = ",", dtype=float):
+        """Very simple python-only function to read in a csv-file from file.
+
+        Args:
+            filepath (str): Full filepath of csv-file to read in.
+            delimiter (str): Delimiter character for separation. Default is ",".
+            dtype: Callable type conversion from string. Default is float.
+
+        Returns:
+            list: Python list of values. Length of the list equals the number of lines.
+        """
+        out = []
+        open_file = open(filepath, "r")
+        for lines in open_file.readlines():
+            string_list = lines.strip().split(delimiter)
+            values_list = [dtype(x.strip()) for x in string_list]
+            out.append(values_list)
+        open_file.close()
+        return out
