@@ -2,14 +2,14 @@ import tensorflow as tf
 # import tensorflow.keras as ks
 from kgcnn.layers.base import GraphBaseLayer
 from kgcnn.layers.gather import GatherNodesIngoing, GatherNodesOutgoing, GatherState
-from kgcnn.layers.keras import Dense, Activation, Concatenate
-from kgcnn.layers.pool.pooling import PoolingNodes
-from kgcnn.layers.pool.pooling import PoolingNodesAttention, PoolingLocalEdgesAttention
+from kgcnn.layers.modules import DenseEmbedding, ActivationEmbedding, LazyConcatenate
+from kgcnn.layers.pooling import PoolingNodes
+from kgcnn.layers.pooling import PoolingNodesAttention, PoolingLocalEdgesAttention
 
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='AttentionHeadGAT')
 class AttentionHeadGAT(GraphBaseLayer):
-    r"""Computes the attention head according to GAT from https://arxiv.org/abs/1710.10903.
+    r"""Computes the attention head according to `GAT <https://arxiv.org/abs/1710.10903>`_ .
     The attention coefficients are computed by :math:`a_{ij} = \sigma(a^T W n_i || W n_j)`,
     optionally by :math:`a_{ij} = \sigma( W n_i || W n_j || e_{ij})` with edges :math:`e_{ij}`.
     The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
@@ -17,8 +17,7 @@ class AttentionHeadGAT(GraphBaseLayer):
     If the graph has no self-loops, they must be added beforehand or use external skip connections.
     And optionally passed through an activation :math:`h_i = \sigma(\sum_j \alpha_{ij} W n_j)`.
 
-    An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+    An edge is defined by index tuple :math:`(i, j)` with the direction of the connection from :math:`j` to :math:`i`.
 
     Args:
         units (int): Units for the linear trafo of node features before attention.
@@ -53,27 +52,24 @@ class AttentionHeadGAT(GraphBaseLayer):
                  **kwargs):
         """Initialize layer."""
         super(AttentionHeadGAT, self).__init__(**kwargs)
-        # graph args
         self.use_edge_features = use_edge_features
         self.use_final_activation = use_final_activation
         self.has_self_loops = has_self_loops
-
-        # dense args
         self.units = int(units)
         self.use_bias = use_bias
-
         kernel_args = {"kernel_regularizer": kernel_regularizer,
                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                        "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
 
-        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args, **self._kgcnn_info)
-        self.lay_alpha = Dense(1, activation=activation, use_bias=False, **kernel_args, **self._kgcnn_info)
-        self.lay_gather_in = GatherNodesIngoing(**self._kgcnn_info)
-        self.lay_gather_out = GatherNodesOutgoing(**self._kgcnn_info)
-        self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
-        self.lay_pool_attention = PoolingLocalEdgesAttention(**self._kgcnn_info)
-        self.lay_final_activ = Activation(activation=activation, **self._kgcnn_info)
+        self.lay_linear_trafo = DenseEmbedding(units, activation="linear", use_bias=use_bias, **kernel_args)
+        self.lay_alpha = DenseEmbedding(1, activation=activation, use_bias=False, **kernel_args)
+        self.lay_gather_in = GatherNodesIngoing()
+        self.lay_gather_out = GatherNodesOutgoing()
+        self.lay_concat = LazyConcatenate(axis=-1)
+        self.lay_pool_attention = PoolingLocalEdgesAttention()
+        if self.use_final_activation:
+            self.lay_final_activ = ActivationEmbedding(activation=activation)
 
     def build(self, input_shape):
         """Build layer."""
@@ -94,18 +90,18 @@ class AttentionHeadGAT(GraphBaseLayer):
         """
         node, edge, edge_index = inputs
 
-        w_n = self.lay_linear_trafo(node)
-        wn_in = self.lay_gather_in([w_n, edge_index])
-        wn_out = self.lay_gather_out([w_n, edge_index])
+        w_n = self.lay_linear_trafo(node, **kwargs)
+        wn_in = self.lay_gather_in([w_n, edge_index], **kwargs)
+        wn_out = self.lay_gather_out([w_n, edge_index], **kwargs)
         if self.use_edge_features:
-            e_ij = self.lay_concat([wn_in, wn_out, edge])
+            e_ij = self.lay_concat([wn_in, wn_out, edge], **kwargs)
         else:
-            e_ij = self.lay_concat([wn_in, wn_out])
-        a_ij = self.lay_alpha(e_ij)  # Should be dimension (batch*None,1)
-        h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
+            e_ij = self.lay_concat([wn_in, wn_out], **kwargs)
+        a_ij = self.lay_alpha(e_ij, **kwargs)  # Should be dimension (batch*None,1)
+        h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)
 
         if self.use_final_activation:
-            h_i = self.lay_final_activ(h_i)
+            h_i = self.lay_final_activ(h_i, **kwargs)
         return h_i
 
     def get_config(self):
@@ -123,7 +119,7 @@ class AttentionHeadGAT(GraphBaseLayer):
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='AttentionHeadGATV2')
 class AttentionHeadGATV2(GraphBaseLayer):
-    r"""Computes the modified attention head according to https://arxiv.org/pdf/2105.14491.pdf.
+    r"""Computes the modified attention head according to `GATv2 <https://arxiv.org/pdf/2105.14491.pdf>`_ .
     The attention coefficients are computed by :math:`a_{ij} = a^T \sigma( W [n_i || n_j] )`,
     optionally by :math:`a_{ij} = a^T \sigma( W [n_i || n_j || e_{ij}] )` with edges :math:`e_{ij}`.
     The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
@@ -131,8 +127,7 @@ class AttentionHeadGATV2(GraphBaseLayer):
     If the graph has no self-loops, they must be added beforehand or use external skip connections.
     And optionally passed through an activation :math:`h_i = \sigma(\sum_j \alpha_{ij} e_{ij})`.
 
-    An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+    An edge is defined by index tuple :math:`(i, j)` with the direction of the connection from :math:`j` to :math:`i`.
 
     Args:
         units (int): Units for the linear trafo of node features before attention.
@@ -167,29 +162,25 @@ class AttentionHeadGATV2(GraphBaseLayer):
                  **kwargs):
         """Initialize layer."""
         super(AttentionHeadGATV2, self).__init__(**kwargs)
-        # graph args
         self.use_edge_features = use_edge_features
         self.use_final_activation = use_final_activation
         self.has_self_loops = has_self_loops
-
-        # dense args
         self.units = int(units)
         self.use_bias = use_bias
-
         kernel_args = {"kernel_regularizer": kernel_regularizer,
                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                        "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
 
-        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args, **self._kgcnn_info)
-        self.lay_alpha_activation = Dense(units, activation=activation, use_bias=use_bias, **kernel_args,
-                                          **self._kgcnn_info)
-        self.lay_alpha = Dense(1, activation="linear", use_bias=False, **kernel_args, **self._kgcnn_info)
-        self.lay_gather_in = GatherNodesIngoing(**self._kgcnn_info)
-        self.lay_gather_out = GatherNodesOutgoing(**self._kgcnn_info)
-        self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
-        self.lay_pool_attention = PoolingLocalEdgesAttention(**self._kgcnn_info)
-        self.lay_final_activ = Activation(activation=activation, **self._kgcnn_info)
+        self.lay_linear_trafo = DenseEmbedding(units, activation="linear", use_bias=use_bias, **kernel_args)
+        self.lay_alpha_activation = DenseEmbedding(units, activation=activation, use_bias=use_bias, **kernel_args)
+        self.lay_alpha = DenseEmbedding(1, activation="linear", use_bias=False, **kernel_args)
+        self.lay_gather_in = GatherNodesIngoing()
+        self.lay_gather_out = GatherNodesOutgoing()
+        self.lay_concat = LazyConcatenate(axis=-1)
+        self.lay_pool_attention = PoolingLocalEdgesAttention()
+        if self.use_final_activation:
+            self.lay_final_activ = ActivationEmbedding(activation=activation)
 
     def build(self, input_shape):
         """Build layer."""
@@ -210,20 +201,20 @@ class AttentionHeadGATV2(GraphBaseLayer):
         """
         node, edge, edge_index = inputs
 
-        w_n = self.lay_linear_trafo(node)
-        n_in = self.lay_gather_in([node, edge_index])
-        n_out = self.lay_gather_out([node, edge_index])
-        wn_out = self.lay_gather_out([w_n, edge_index])
+        w_n = self.lay_linear_trafo(node, **kwargs)
+        n_in = self.lay_gather_in([node, edge_index], **kwargs)
+        n_out = self.lay_gather_out([node, edge_index], **kwargs)
+        wn_out = self.lay_gather_out([w_n, edge_index], **kwargs)
         if self.use_edge_features:
-            e_ij = self.lay_concat([n_in, n_out, edge])
+            e_ij = self.lay_concat([n_in, n_out, edge], **kwargs)
         else:
-            e_ij = self.lay_concat([n_in, n_out])
-        a_ij = self.lay_alpha_activation(e_ij)
-        a_ij = self.lay_alpha(a_ij)
-        h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
+            e_ij = self.lay_concat([n_in, n_out], **kwargs)
+        a_ij = self.lay_alpha_activation(e_ij, **kwargs)
+        a_ij = self.lay_alpha(a_ij, **kwargs)
+        h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)
 
         if self.use_final_activation:
-            h_i = self.lay_final_activ(h_i)
+            h_i = self.lay_final_activ(h_i, **kwargs)
         return h_i
 
     def get_config(self):
@@ -241,13 +232,13 @@ class AttentionHeadGATV2(GraphBaseLayer):
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='AttentiveHeadFP')
 class AttentiveHeadFP(GraphBaseLayer):
-    r"""Computes the attention head for Attentive FP model.
+    r"""Computes the attention head for `Attentive FP <https://doi.org/10.1021/acs.jmedchem.9b00959>`_ model.
     The attention coefficients are computed by :math:`a_{ij} = \sigma_1( W_1 [h_i || h_j] )`.
     The initial representation :math:`h_i` and :math:`h_j` must be calculated beforehand.
     The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
     And finally pooled for context :math:`C_i = \sigma_2(\sum_j \alpha_{ij} W_2 h_j)`.
 
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+    An edge is defined by index tuple :math:`(i, j)` with the direction of the connection from :math:`j` to :math:`i`.
 
     Args:
         units (int): Units for the linear trafo of node features before attention.
@@ -280,31 +271,26 @@ class AttentiveHeadFP(GraphBaseLayer):
                  **kwargs):
         """Initialize layer."""
         super(AttentiveHeadFP, self).__init__(**kwargs)
-        # graph args
         self.use_edge_features = use_edge_features
-
-        # dense args
         self.units = int(units)
         self.use_bias = use_bias
-
         kernel_args = {"kernel_regularizer": kernel_regularizer,
                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                        "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
 
-        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args, **self._kgcnn_info)
-        self.lay_alpha_activation = Dense(units, activation=activation, use_bias=use_bias, **kernel_args,
-                                          **self._kgcnn_info)
-        self.lay_alpha = Dense(1, activation="linear", use_bias=False, **kernel_args, **self._kgcnn_info)
-        self.lay_gather_in = GatherNodesIngoing(**self._kgcnn_info)
-        self.lay_gather_out = GatherNodesOutgoing(**self._kgcnn_info)
-        self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
-        self.lay_pool_attention = PoolingLocalEdgesAttention(**self._kgcnn_info)
-        self.lay_final_activ = Activation(activation=activation_context, **self._kgcnn_info)
+        self.lay_linear_trafo = DenseEmbedding(units, activation="linear", use_bias=use_bias, **kernel_args)
+        self.lay_alpha_activation = DenseEmbedding(units, activation=activation, use_bias=use_bias, **kernel_args)
+        self.lay_alpha = DenseEmbedding(1, activation="linear", use_bias=False, **kernel_args)
+        self.lay_gather_in = GatherNodesIngoing()
+        self.lay_gather_out = GatherNodesOutgoing()
+        self.lay_concat = LazyConcatenate(axis=-1)
+        self.lay_pool_attention = PoolingLocalEdgesAttention()
+        self.lay_final_activ = ActivationEmbedding(activation=activation_context)
         if use_edge_features:
-            self.lay_fc1 = Dense(units, activation=activation, use_bias=use_bias, **kernel_args, **self._kgcnn_info)
-            self.lay_fc2 = Dense(units, activation=activation, use_bias=use_bias, **kernel_args, **self._kgcnn_info)
-            self.lay_concat_edge = Concatenate(axis=-1, **self._kgcnn_info)
+            self.lay_fc1 = DenseEmbedding(units, activation=activation, use_bias=use_bias, **kernel_args)
+            self.lay_fc2 = DenseEmbedding(units, activation=activation, use_bias=use_bias, **kernel_args)
+            self.lay_concat_edge = LazyConcatenate(axis=-1)
 
     def build(self, input_shape):
         """Build layer."""
@@ -326,22 +312,22 @@ class AttentiveHeadFP(GraphBaseLayer):
         node, edge, edge_index = inputs
 
         if self.use_edge_features:
-            n_in = self.lay_gather_in([node, edge_index])
-            n_out = self.lay_gather_out([node, edge_index])
-            n_in = self.lay_fc1(n_in)
-            n_out = self.lay_concat_edge([n_out, edge])
-            n_out = self.lay_fc2(n_out)
+            n_in = self.lay_gather_in([node, edge_index], **kwargs)
+            n_out = self.lay_gather_out([node, edge_index], **kwargs)
+            n_in = self.lay_fc1(n_in, **kwargs)
+            n_out = self.lay_concat_edge([n_out, edge], **kwargs)
+            n_out = self.lay_fc2(n_out, **kwargs)
         else:
-            n_in = self.lay_gather_in([node, edge_index])
-            n_out = self.lay_gather_out([node, edge_index])
+            n_in = self.lay_gather_in([node, edge_index], **kwargs)
+            n_out = self.lay_gather_out([node, edge_index], **kwargs)
 
-        wn_out = self.lay_linear_trafo(n_out)
-        e_ij = self.lay_concat([n_in, n_out])
-        e_ij = self.lay_alpha_activation(e_ij)  # Maybe uses GAT original definition.
+        wn_out = self.lay_linear_trafo(n_out, **kwargs)
+        e_ij = self.lay_concat([n_in, n_out], **kwargs)
+        e_ij = self.lay_alpha_activation(e_ij, **kwargs)  # Maybe uses GAT original definition.
         # a_ij = e_ij
-        a_ij = self.lay_alpha(e_ij)  # Should be dimension (batch,None,1) not fully clear in original paper SI.
-        n_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
-        out = self.lay_final_activ(n_i)
+        a_ij = self.lay_alpha(e_ij, **kwargs)  # Should be dimension (batch,None,1) not fully clear in original paper.
+        n_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)
+        out = self.lay_final_activ(n_i, **kwargs)
         return out
 
     def get_config(self):
@@ -360,7 +346,8 @@ class AttentiveHeadFP(GraphBaseLayer):
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingNodesAttentive')
 class PoolingNodesAttentive(GraphBaseLayer):
-    r"""Computes the attentive pooling for node embeddings.
+    r"""Computes the attentive pooling for node embeddings for
+    `Attentive FP <https://doi.org/10.1021/acs.jmedchem.9b00959>`_ model.
 
     Args:
         units (int): Units for the linear trafo of node features before attention.
@@ -404,9 +391,7 @@ class PoolingNodesAttentive(GraphBaseLayer):
         super(PoolingNodesAttentive, self).__init__(**kwargs)
         self.pooling_method = pooling_method
         self.depth = depth
-        # dense args
         self.units = int(units)
-
         kernel_args = {"use_bias": use_bias, "kernel_regularizer": kernel_regularizer,
                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
@@ -419,13 +404,13 @@ class PoolingNodesAttentive(GraphBaseLayer):
                     "recurrent_constraint": recurrent_constraint, "bias_constraint": bias_constraint,
                     "dropout": dropout, "recurrent_dropout": recurrent_dropout, "reset_after": reset_after}
 
-        self.lay_linear_trafo = Dense(units, activation="linear", **kernel_args, **self._kgcnn_info)
-        self.lay_alpha = Dense(1, activation=activation, **kernel_args, **self._kgcnn_info)
-        self.lay_gather_s = GatherState(**self._kgcnn_info)
-        self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
-        self.lay_pool_start = PoolingNodes(pooling_method=self.pooling_method, **self._kgcnn_info)
-        self.lay_pool_attention = PoolingNodesAttention(**self._kgcnn_info)
-        self.lay_final_activ = Activation(activation=activation_context, **self._kgcnn_info)
+        self.lay_linear_trafo = DenseEmbedding(units, activation="linear", **kernel_args)
+        self.lay_alpha = DenseEmbedding(1, activation=activation, **kernel_args)
+        self.lay_gather_s = GatherState()
+        self.lay_concat = LazyConcatenate(axis=-1)
+        self.lay_pool_start = PoolingNodes(pooling_method=self.pooling_method)
+        self.lay_pool_attention = PoolingNodesAttention()
+        self.lay_final_activ = ActivationEmbedding(activation=activation_context)
         self.lay_gru = tf.keras.layers.GRUCell(units=units, activation="tanh", **gru_args)
 
     def build(self, input_shape):
@@ -445,14 +430,14 @@ class PoolingNodesAttentive(GraphBaseLayer):
         """
         node = inputs
 
-        h = self.lay_pool_start(node)
-        wn = self.lay_linear_trafo(node)
+        h = self.lay_pool_start(node, **kwargs)
+        wn = self.lay_linear_trafo(node, **kwargs)
         for _ in range(self.depth):
-            hv = self.lay_gather_s([h, node])
-            ev = self.lay_concat([hv, node])
-            av = self.lay_alpha(ev)
-            cont = self.lay_pool_attention([wn, av])
-            cont = self.lay_final_activ(cont)
+            hv = self.lay_gather_s([h, node], **kwargs)
+            ev = self.lay_concat([hv, node], **kwargs)
+            av = self.lay_alpha(ev, **kwargs)
+            cont = self.lay_pool_attention([wn, av], **kwargs)
+            cont = self.lay_final_activ(cont, **kwargs)
             h, _ = self.lay_gru(cont, h, **kwargs)
 
         out = h
