@@ -5,29 +5,33 @@ from kgcnn.ops.partition import partition_row_indexing
 from kgcnn.ops.axis import get_positive_axis
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherNodes')
-class GatherNodes(GraphBaseLayer):
-    """Gather nodes by indices, e.g. that define an edge.
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherEmbedding')
+class GatherEmbedding(GraphBaseLayer):
+    r"""Gather node or edge embedding by index list, e.g. that define an edge. The embeddings for multiple indices
+    are concatenated.
 
-    An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+    An edge is defined by index tuple :math:`(i,j)`.
+    In the default definition for this layer index :math:`i` is expected to be the
+    receiving or target node (in standard case of directed edges).
+
+    Args:
+        axis (int): The axis to gather embeddings from. Default is 1.
+        concat_axis (int): The axis which concatenates embeddings. Default is 2.
     """
 
     def __init__(self,
-                 axis=1,
-                 concat_axis=2,
+                 axis: int = 1,
+                 concat_axis: int = 2,
                  **kwargs):
         """Initialize layer."""
-        super(GatherNodes, self).__init__(**kwargs)
+        super(GatherEmbedding, self).__init__(**kwargs)
         self.concat_axis = concat_axis
         self.axis = axis
 
     def build(self, input_shape):
-        super(GatherNodes, self).build(input_shape)
+        super(GatherEmbedding, self).build(input_shape)
         if len(input_shape) != 2:
             print("WARNING: Number of inputs for layer", self.name, "is expected to be 2.")
-        if [i for i in range(len(input_shape[0]))][self.axis] == 0:
-            print("WARNING: Shape error for", self.name, ", gather from batch-dimension is not intended.")
 
     def call(self, inputs, **kwargs):
         """Forward pass.
@@ -53,12 +57,12 @@ class GatherNodes(GraphBaseLayer):
                                                        from_indexing=self.node_indexing)
                 out = tf.gather(node, disjoint_list, axis=0)
                 if self.concat_axis == 2:
-                    out = tf.keras.backend.concatenate([out[:, i] for i in range(edge_index.shape[-1])], axis=1)
+                    out = tf.concat([out[:, i] for i in range(edge_index.shape[-1])], axis=1)
                 out = tf.RaggedTensor.from_row_lengths(out, edge_part, validate=self.ragged_validate)
                 return out
 
         # For arbitrary gather from ragged tensor use tf.gather with batch_dims=1.
-        # Works in tf.__version__>=2.4
+        # Works in tf.__version__>=2.4 now!
         out = tf.gather(inputs[0], inputs[1], batch_dims=1, axis=self.axis)
         if self.concat_axis is not None:
             out = tf.concat([tf.gather(out, i, axis=self.concat_axis) for i in range(out.shape[self.concat_axis])],
@@ -66,21 +70,32 @@ class GatherNodes(GraphBaseLayer):
         return out
 
     def get_config(self):
-        config = super(GatherNodes, self).get_config()
+        config = super(GatherEmbedding, self).get_config()
         config.update({"concat_axis": self.concat_axis, "axis": self.axis})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherNodesSelection')
-class GatherNodesSelection(GraphBaseLayer):
-    """Gather nodes with a defined index. E.g. for ingoing or outgoing nodes or angles.
+GatherNodes = GatherEmbedding
 
-    For ingoing nodes, layer uses only index[0]. An edge is defined by index tuple (i,j) with i<-j connection.
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherEmbeddingSelection')
+class GatherEmbeddingSelection(GraphBaseLayer):
+    r"""Gather node or edge embedding for a defined index in the index list.
+    E.g. for ingoing or outgoing nodes or angles. Returns a list of embeddings for each :obj:`selection_index`.
+
+    An edge is defined by index tuple :math:`(i,j)`.
+    In the default definition for this layer index :math:`i` is expected to be the
+    receiving or target node (in standard case of directed edges).
+
+    Args:
+        selection_index (list): Which indices to gather embeddings for.
+        axis (int): Axis to gather embeddings from. Default is 1.
+        axis_indices (int): From which axis to take the indices for gather. Default is 2.
     """
 
-    def __init__(self, selection_index, axis=1, axis_indices=2, **kwargs):
+    def __init__(self, selection_index, axis: int = 1, axis_indices: int = 2, **kwargs):
         """Initialize layer."""
-        super(GatherNodesSelection, self).__init__(**kwargs)
+        super(GatherEmbeddingSelection, self).__init__(**kwargs)
         self.axis = axis
         self.axis_indices = axis_indices
 
@@ -94,11 +109,11 @@ class GatherNodesSelection(GraphBaseLayer):
 
     def build(self, input_shape):
         """Build layer."""
-        super(GatherNodesSelection, self).build(input_shape)
+        super(GatherEmbeddingSelection, self).build(input_shape)
         if len(input_shape) != 2:
-            print("WARNING: Number of inputs for layer", self.name, "is expected to be 2.")
+            print("WARNING:kgcnn: Number of inputs for layer", self.name, "is expected to be 2.")
         if [i for i in range(len(input_shape[0]))][self.axis] == 0:
-            print("WARNING: Shape error for", self.name, ", gather from batch-dimension is not intended.")
+            print("WARNING:kgcnn: Shape error for", self.name, ", gather from batch-dimension is not intended.")
 
     def call(self, inputs, **kwargs):
         """Forward pass.
@@ -134,20 +149,27 @@ class GatherNodesSelection(GraphBaseLayer):
         return out
 
     def get_config(self):
-        config = super(GatherNodesSelection, self).get_config()
+        config = super(GatherEmbeddingSelection, self).get_config()
         config.update({"axis": self.axis, "axis_indices": self.axis_indices, "selection_index": self.selection_index})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherNodesOutgoing')
-class GatherNodesOutgoing(GatherNodesSelection):
-    """Gather nodes by indices.
+GatherNodesSelection = GatherEmbeddingSelection
 
-    For outgoing nodes, layer uses only index[1]. An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherNodesOutgoing')
+class GatherNodesOutgoing(GatherEmbeddingSelection):
+    r"""Gather nodes by index :math:`j`, here defined as sending or outgoing.
+
+    An edge is defined by index tuple :math:`(i,j)`.
+    In the default definition for this layer index :math:`i` is expected to be the
+    receiving or target node (in standard case of directed edges).
+
+    Args:
+        selection_index (list): Which index to gather embeddings for. Default is 1.
     """
 
-    def __init__(self, selection_index=1, **kwargs):
+    def __init__(self, selection_index: int = 1, **kwargs):
         super(GatherNodesOutgoing, self).__init__(selection_index=selection_index, **kwargs)
 
     def call(self, inputs, **kwargs):
@@ -155,14 +177,18 @@ class GatherNodesOutgoing(GatherNodesSelection):
 
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherNodesIngoing')
-class GatherNodesIngoing(GatherNodesSelection):
-    """Gather nodes by edge edge_indices.
-    
-    For ingoing nodes, layer uses only index[0]. An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+class GatherNodesIngoing(GatherEmbeddingSelection):
+    r"""Gather nodes by index :math:`i`, here defined as receiving or ingoing.
+
+    An edge is defined by index tuple :math:`(i,j)`.
+    In the default definition for this layer index :math:`i` is expected to be the
+    receiving or target node (in standard case of directed edges).
+
+    Args:
+        selection_index (list): Which index to gather embeddings for. Default is 0.
     """
 
-    def __init__(self, selection_index=0, **kwargs):
+    def __init__(self, selection_index: int = 0, **kwargs):
         super(GatherNodesIngoing, self).__init__(selection_index=selection_index, **kwargs)
 
     def call(self, inputs, **kwargs):
@@ -171,7 +197,7 @@ class GatherNodesIngoing(GatherNodesSelection):
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='GatherState')
 class GatherState(GraphBaseLayer):
-    """Layer to repeat environment or global state for node or edge lists.
+    """Layer to repeat environment or global state for a specific node or edge list.
     
     To repeat the correct environment for each sample, a tensor with the target length/partition is required.
     """
@@ -196,11 +222,13 @@ class GatherState(GraphBaseLayer):
         Returns:
             tf.RaggedTensor: Graph embedding with repeated single state for each graph of shape (batch, [N], F).
         """
+        env = inputs[0]
         dyn_inputs = inputs[1]
 
-        # We cast to values here
-        env = inputs[0]
-        target_len = dyn_inputs.row_lengths()
+        if isinstance(dyn_inputs, tf.RaggedTensor):
+            target_len = dyn_inputs.row_lengths()
+        else:
+            target_len = tf.repeat(tf.shape(dyn_inputs)[1], tf.shape(dyn_inputs)[0])
 
         out = tf.repeat(env, target_len, axis=0)
         out = tf.RaggedTensor.from_row_lengths(out, target_len, validate=self.ragged_validate)
