@@ -8,20 +8,26 @@ from kgcnn.ops.segment import segment_ops_by_name, segment_softmax
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdges')
 class PoolingLocalEdges(GraphBaseLayer):
-    """Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
-    
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
-    Apply e.g. segment_mean for index[0] incoming nodes. 
-    Important: tensor_index[:,0] are sorted for segment-operation.
+    r"""The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
+    corresponding to the receiving node, which is defined by edge indices.
+    The term pooling is here used as aggregating rather than reducing the graph as in graph pooling.
+
+    Apply e.g. sum or mean on edges with same target ID taken from the (edge) index_tensor, that has a list of
+    all connections as :math:`(i, j)`. In the default definition for this layer index :math:`i` is expected ot be the
+    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index`.
+
+    Note: index_tensor[:, :, pooling_index] is sorted for the subsequent segment-operation.
     
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
+        pooling_index (int): Index from edge_indices to pick ID's for pooling edge-like embeddings. Default is 0.
     """
 
-    def __init__(self, pooling_method="mean", **kwargs):
+    def __init__(self, pooling_method="mean", pooling_index=0, **kwargs):
         """Initialize layer."""
         super(PoolingLocalEdges, self).__init__(**kwargs)
         self.pooling_method = pooling_method
+        self.pooling_index = pooling_index
 
     def build(self, input_shape):
         """Build layer."""
@@ -40,11 +46,12 @@ class PoolingLocalEdges(GraphBaseLayer):
         Returns:
             tf.RaggedTensor: Pooled feature tensor of pooled edge features for each node.
         """
-        dyn_inputs = inputs
-        # We cast to values here
-        nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
-        edge, _ = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
-        edgeind, edge_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
+        # Need ragged input but can be generalized in the future.
+        self.assert_ragged_input_rank(inputs)
+
+        nod, node_part = inputs[0].values, inputs[0].row_splits
+        edge, _ = inputs[1].values, inputs[1].row_lengths()
+        edgeind, edge_part = inputs[2].values, inputs[2].row_lengths()
 
         shiftind = partition_row_indexing(edgeind, node_part, edge_part,
                                           partition_type_target="row_splits",
@@ -52,15 +59,17 @@ class PoolingLocalEdges(GraphBaseLayer):
                                           to_indexing='batch',
                                           from_indexing=self.node_indexing)
 
-        nodind = shiftind[:, 0]  # Pick first index eg. ingoing
+        nodind = shiftind[:, self.pooling_index]  # Pick index eg. ingoing
         dens = edge
         if not self.is_sorted:
-            # Sort edgeindices
             node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
             nodind = tf.gather(nodind, node_order, axis=0)
             dens = tf.gather(dens, node_order, axis=0)
+
         # Pooling via e.g. segment_sum
         out = segment_ops_by_name(self.pooling_method, dens, nodind)
+
+        # If not unsort_segment operation need a scatter here.
         if self.has_unconnected:
             out = tf.scatter_nd(tf.keras.backend.expand_dims(tf.range(tf.shape(out)[0]), axis=-1), out,
                                 tf.concat([tf.shape(nod)[:1], tf.shape(out)[1:]], axis=0))
@@ -71,7 +80,7 @@ class PoolingLocalEdges(GraphBaseLayer):
     def get_config(self):
         """Update layer config."""
         config = super(PoolingLocalEdges, self).get_config()
-        config.update({"pooling_method": self.pooling_method})
+        config.update({"pooling_method": self.pooling_method, "pooling_index": self.pooling_index})
         return config
 
 
@@ -80,24 +89,33 @@ PoolingLocalMessages = PoolingLocalEdges  # For now they are synonyms
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedLocalEdges')
 class PoolingWeightedLocalEdges(GraphBaseLayer):
-    """Pooling all edges or message/edge-like features per node, corresponding to node assigned by edge_indices.
-    
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
-    Apply e.g. segment_mean for index[0] incoming nodes. 
-    Important: tensor_index[:,0] could be sorted for segment-operation.
-    
+    r"""The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
+    corresponding to the receiving node, which is defined by edge indices.
+    The term pooling is here used as aggregating rather than reducing the graph as in graph pooling.
+
+    Apply e.g. sum or mean on edges with same target ID taken from the (edge) index-tensor, that has a list of
+    all connections as :math:`(i, j)`. In the default definition for this layer index :math:`i` is expected ot be the
+    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index`.
+
+    .. note::
+        In addition of aggregating edge embeddings a weight tensor must be supplied that scales each edge before
+        pooling. Must broadcast.
+
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
         normalize_by_weights (bool): Normalize the pooled output by the sum of weights. Default is False.
+        pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
     """
 
     def __init__(self, pooling_method="mean",
                  normalize_by_weights=False,
+                 pooling_index=0,
                  **kwargs):
         """Initialize layer."""
         super(PoolingWeightedLocalEdges, self).__init__(**kwargs)
         self.pooling_method = pooling_method
         self.normalize_by_weights = normalize_by_weights
+        self.pooling_index = pooling_index
 
     def build(self, input_shape):
         """Build layer."""
@@ -118,12 +136,12 @@ class PoolingWeightedLocalEdges(GraphBaseLayer):
         Returns:
             tf.RaggedTensor: Pooled feature tensor of pooled edge features for each node of shape (batch, [N], F)
         """
-        dyn_inputs = inputs
-        # We cast to values here
-        nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
-        edge, _ = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
-        edgeind, edge_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
-        weights, _ = dyn_inputs[3].values, dyn_inputs[3].row_lengths()
+        self.assert_ragged_input_rank(inputs)
+
+        nod, node_part = inputs[0].values, inputs[0].row_splits
+        edge, _ = inputs[1].values, inputs[1].row_lengths()
+        edgeind, edge_part = inputs[2].values, inputs[2].row_lengths()
+        weights, _ = inputs[3].values, inputs[3].row_lengths()
 
         shiftind = partition_row_indexing(edgeind, node_part, edge_part,
                                           partition_type_target="row_splits", partition_type_index="row_length",
@@ -131,10 +149,9 @@ class PoolingWeightedLocalEdges(GraphBaseLayer):
 
         wval = weights
         dens = edge * wval
-        nodind = shiftind[:, 0]
+        nodind = shiftind[:, self.pooling_index]
 
         if not self.is_sorted:
-            # Sort edgeindices
             node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
             nodind = tf.gather(nodind, node_order, axis=0)
             dens = tf.gather(dens, node_order, axis=0)
@@ -156,16 +173,18 @@ class PoolingWeightedLocalEdges(GraphBaseLayer):
     def get_config(self):
         """Update layer config."""
         config = super(PoolingWeightedLocalEdges, self).get_config()
-        config.update({"pooling_method": self.pooling_method, "normalize_by_weights": self.normalize_by_weights})
+        config.update({"pooling_method": self.pooling_method, "normalize_by_weights": self.normalize_by_weights,
+                       "pooling_index": self.pooling_index})
         return config
 
 
 PoolingWeightedLocalMessages = PoolingWeightedLocalEdges  # For now they are synonyms
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingNodes')
-class PoolingNodes(GraphBaseLayer):
-    """Polling all nodes per batch. The batch assignment is given by a length-tensor.
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingEmbedding')
+class PoolingEmbedding(GraphBaseLayer):
+    """Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
+    ::obj`tf.Tensor`.
     
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -173,41 +192,50 @@ class PoolingNodes(GraphBaseLayer):
 
     def __init__(self, pooling_method="mean", **kwargs):
         """Initialize layer."""
-        super(PoolingNodes, self).__init__(**kwargs)
+        super(PoolingEmbedding, self).__init__(**kwargs)
         self.pooling_method = pooling_method
 
     def build(self, input_shape):
         """Build layer."""
-        super(PoolingNodes, self).build(input_shape)
+        super(PoolingEmbedding, self).build(input_shape)
 
     def call(self, inputs, **kwargs):
         """Forward pass.
 
         Args:
-            inputs (tf.RaggedTensor): Node features of shape (batch, [N], F)
+            inputs (tf.RaggedTensor): Embedding tensor of shape (batch, [N], F)
     
         Returns:
             tf.Tensor: Pooled node features of shape (batch, F)
         """
-        dyn_inputs = [inputs]
+        # Need ragged input but can be generalized in the future.
+        self.assert_ragged_input_rank(inputs)
         # We cast to values here
-        nod, batchi = dyn_inputs[0].values, dyn_inputs[0].value_rowids()
+        nod, batchi = inputs.values, inputs.value_rowids()
 
+        # Could also use reduce_sum here.
         out = segment_ops_by_name(self.pooling_method, nod, batchi)
-
-        # Output should have correct shape
         return out
 
     def get_config(self):
         """Update layer config."""
-        config = super(PoolingNodes, self).get_config()
+        config = super(PoolingEmbedding, self).get_config()
         config.update({"pooling_method": self.pooling_method})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedNodes')
-class PoolingWeightedNodes(GraphBaseLayer):
-    """Polling all nodes per batch. The batch assignment is given by a length-tensor.
+PoolingNodes = PoolingEmbedding
+PoolingGlobalEdges = PoolingEmbedding
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedEmbedding')
+class PoolingWeightedEmbedding(GraphBaseLayer):
+    """Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
+    ::obj`tf.Tensor`.
+
+    .. note::
+        In addition of pooling embeddings a weight tensor must be supplied that scales each embedding before
+        pooling. Must broadcast.
 
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -215,12 +243,12 @@ class PoolingWeightedNodes(GraphBaseLayer):
 
     def __init__(self, pooling_method="mean", **kwargs):
         """Initialize layer."""
-        super(PoolingWeightedNodes, self).__init__(**kwargs)
+        super(PoolingWeightedEmbedding, self).__init__(**kwargs)
         self.pooling_method = pooling_method
 
     def build(self, input_shape):
         """Build layer."""
-        super(PoolingWeightedNodes, self).build(input_shape)
+        super(PoolingWeightedEmbedding, self).build(input_shape)
 
     def call(self, inputs, **kwargs):
         """Forward pass.
@@ -234,77 +262,41 @@ class PoolingWeightedNodes(GraphBaseLayer):
         Returns:
             tf.Tensor: Pooled node features of shape (batch, F)
         """
-        dyn_inputs = inputs
+        # Need ragged input but can be generalized in the future.
+        self.assert_ragged_input_rank(inputs)
         # We cast to values here
-        nod, batchi = dyn_inputs[0].values, dyn_inputs[0].value_rowids()
-        weights, _ = dyn_inputs[1].values, dyn_inputs[1].value_rowids()
-
+        nod, batchi = inputs[0].values, inputs[0].value_rowids()
+        weights, _ = inputs[1].values, inputs[1].value_rowids()
         nod = tf.math.multiply(nod, weights)
+        # Could also use reduce_sum here.
         out = segment_ops_by_name(self.pooling_method, nod, batchi)
-        # Output should have correct shape
         return out
 
     def get_config(self):
         """Update layer config."""
-        config = super(PoolingWeightedNodes, self).get_config()
+        config = super(PoolingWeightedEmbedding, self).get_config()
         config.update({"pooling_method": self.pooling_method})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingGlobalEdges')
-class PoolingGlobalEdges(GraphBaseLayer):
-    """Pooling all edges per graph. The batch assignment is given by a length-tensor.
-
-    Args:
-        pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
-    """
-
-    def __init__(self, pooling_method="mean", **kwargs):
-        """Initialize layer."""
-        super(PoolingGlobalEdges, self).__init__(**kwargs)
-        self.pooling_method = pooling_method
-
-    def build(self, input_shape):
-        """Build layer."""
-        super(PoolingGlobalEdges, self).build(input_shape)
-
-    def call(self, inputs, **kwargs):
-        """Forward pass.
-
-        Args:
-            inputs (tf.RaggedTensor): Edge features or message embeddings of shape (batch, [M], F)
-    
-        Returns:
-            tf.Tensor: Pooled edges feature list of shape (batch, F).
-        """
-        dyn_inputs = [inputs]
-        # We cast to values here
-        edge, batchi = dyn_inputs[0].values, dyn_inputs[0].value_rowids()
-
-        out = segment_ops_by_name(self.pooling_method, edge, batchi)
-        # Output already has correct shape and type
-        return out
-
-    def get_config(self):
-        """Update layer config."""
-        config = super(PoolingGlobalEdges, self).get_config()
-        config.update({"pooling_method": self.pooling_method})
-        return config
+PoolingWeightedNodes = PoolingWeightedEmbedding
+PoolingWeightedGlobalEdges = PoolingWeightedEmbedding
 
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdgesLSTM')
 class PoolingLocalEdgesLSTM(GraphBaseLayer):
-    """
-    Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
-    Uses LSTM to aggregate Node-features.
+    """The main aggregation or pooling layer to collect all edges or edge-like embeddings per node,
+    corresponding to the receiving node, which is defined by edge indices.
+    The term pooling is here used as aggregating rather than reducing the graph as in graph pooling.
 
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
-    Apply e.g. segment_mean for index[0] incoming nodes.
-    Important: tensor_index[:,0] are sorted for segment-operation.
+    Here, apply LSTM on edges with same target ID taken from the (edge) index_tensor, that has a list of
+    all connections as :math:`(i, j)`. In the default definition for this layer index :math:`i` is expected ot be the
+    receiving or target node (in standard case of directed edges). This can be changed by setting :obj:`pooling_index`.
 
     Args:
         units (int): Units for LSTM cell.
         pooling_method (str): Pooling method. Default is 'LSTM', is ignored.
+        pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
         activation: Activation function to use.
             Default: hyperbolic tangent (`tanh`). If you pass `None`, no activation
             is applied (ie. "linear" activation: `a(x) = x`).
@@ -366,6 +358,7 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
     def __init__(self,
                  units,
                  pooling_method="LSTM",
+                 pooling_index=0,
                  activation='tanh', recurrent_activation='sigmoid',
                  use_bias=True, kernel_initializer='glorot_uniform',
                  recurrent_initializer='orthogonal',
@@ -379,6 +372,7 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
         """Initialize layer."""
         super(PoolingLocalEdgesLSTM, self).__init__(**kwargs)
         self.pooling_method = pooling_method
+        self.pooling_index = pooling_index
 
         self.lstm_unit = ks.layers.LSTM(units=units, activation=activation, recurrent_activation=recurrent_activation,
                                         use_bias=use_bias, kernel_initializer=kernel_initializer,
@@ -413,11 +407,11 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
         Returns:
             tf.RaggedTensor: Feature tensor of pooled edge features for each node of shape (batch, [N], F)
         """
-        dyn_inputs = inputs
-        # We cast to values here
-        nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
-        edge, _ = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
-        edgeind, edge_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
+        self.assert_ragged_input_rank(inputs)
+
+        nod, node_part = inputs[0].values, inputs[0].row_splits
+        edge, _ = inputs[1].values, inputs[1].row_lengths()
+        edgeind, edge_part = inputs[2].values, inputs[2].row_lengths()
 
         shiftind = partition_row_indexing(edgeind, node_part, edge_part,
                                           partition_type_target="row_splits",
@@ -425,7 +419,7 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
                                           to_indexing='batch',
                                           from_indexing=self.node_indexing)
 
-        nodind = shiftind[:, 0]  # Pick first index eg. ingoing
+        nodind = shiftind[:, self.pooling_index]  # Pick first index eg. ingoing
         dens = edge
         if not self.is_sorted:
             # Sort edgeindices
@@ -450,7 +444,7 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
     def get_config(self):
         """Update layer config."""
         config = super(PoolingLocalEdgesLSTM, self).get_config()
-        config.update({"pooling_method": self.pooling_method})
+        config.update({"pooling_method": self.pooling_method, "pooling_index": self.pooling_index})
         conf_lstm = self.lstm_unit.get_config()
         lstm_param = ["activation", "recurrent_activation", "use_bias", "kernel_initializer", "recurrent_initializer",
                       "bias_initializer", "unit_forget_bias", "kernel_regularizer", "recurrent_regularizer",
@@ -467,7 +461,8 @@ PoolingLocalMessagesLSTM = PoolingLocalEdgesLSTM  # For now they are synonyms
 
 @tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdgesAttention')
 class PoolingLocalEdgesAttention(GraphBaseLayer):
-    r"""Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
+    r"""Pooling or aggregation of all edges or edge-like features per node,
+    corresponding to node assigned by edge indices.
     Uses attention for pooling. i.e. :math:`n_i =  \sum_j \alpha_{ij} e_{ij}`
     The attention is computed via: :math:`\alpha_ij = \text{softmax}_j (a_{ij})` from the
     attention coefficients :math:`a_{ij}`.
@@ -475,14 +470,19 @@ class PoolingLocalEdgesAttention(GraphBaseLayer):
     are passed to this layer as input. Thereby this layer has no weights and only does pooling.
     In summary, :math:`n_i = \sum_j \text{softmax}_j (a_{ij}) e_{ij}` is computed by the layer.
 
-    An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
-    Important: tensor_index[:,0] are sorted for segment-operation for pooling with respect to tensor_index[:,0].
+    An edge is defined by index tuple :math:`(i, j)` with :math:`i` being the receiving node in the default definition,
+    but can be changed by pooling_index.
+    Important: ID's for segment-operation and for pooling of edges are taken from edge-index-tensor.
+    They are sorted for faster pooling from tensor_index[:, :, pooling_index].
+
+    Args:
+        pooling_index (int): Index to pick ID's for pooling edge-like embeddings. Default is 0.
     """
 
-    def __init__(self, **kwargs):
+    def __init__(self, pooling_index=0, **kwargs):
         """Initialize layer."""
         super(PoolingLocalEdgesAttention, self).__init__(**kwargs)
+        self.pooling_index = pooling_index
 
     def build(self, input_shape):
         """Build layer."""
@@ -502,19 +502,19 @@ class PoolingLocalEdgesAttention(GraphBaseLayer):
         Returns:
             tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node of shape (batch, [N], F)
         """
-        dyn_inputs = inputs
-
+        # Need ragged input but can be generalized in the future.
+        self.assert_ragged_input_rank(inputs)
         # We cast to values here
-        nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_lengths()
-        edge = dyn_inputs[1].values
-        attention = dyn_inputs[2].values
-        edgeind, edge_part = dyn_inputs[3].values, dyn_inputs[3].row_lengths()
+        nod, node_part = inputs[0].values, inputs[0].row_lengths()
+        edge = inputs[1].values
+        attention = inputs[2].values
+        edgeind, edge_part = inputs[3].values, inputs[3].row_lengths()
 
         shiftind = partition_row_indexing(edgeind, node_part, edge_part, partition_type_target="row_length",
                                           partition_type_index="row_length", to_indexing='batch',
                                           from_indexing=self.node_indexing)
 
-        nodind = shiftind[:, 0]  # Pick first index eg. ingoing
+        nodind = shiftind[:, self.pooling_index]  # Pick first index eg. ingoing
         dens = edge
         ats = attention
         if not self.is_sorted:
@@ -541,27 +541,29 @@ class PoolingLocalEdgesAttention(GraphBaseLayer):
     def get_config(self):
         """Update layer config."""
         config = super(PoolingLocalEdgesAttention, self).get_config()
+        config.update({"pooling_index": self.pooling_index})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingNodesAttention')
-class PoolingNodesAttention(GraphBaseLayer):
-    r"""Pooling all nodes.
-    Uses attention for pooling. i.e. :math:`s =  \sum_j \alpha_{i} n_i`
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingEmbeddingAttention')
+class PoolingEmbeddingAttention(GraphBaseLayer):
+    r"""Polling all embeddings of edges or nodes per batch to obtain a graph level embedding in form of a
+    ::obj`tf.Tensor`.
+
+    Uses attention for pooling. i.e. :math:`s =  \sum_j \alpha_{i} n_i`.
     The attention is computed via: :math:`\alpha_i = \text{softmax}_i(a_i)` from the attention coefficients :math:`a_i`.
     The attention coefficients must be computed beforehand by edge features or by :math:`\sigma( W [s || n_i])` and
     are passed to this layer as input. Thereby this layer has no weights and only does pooling.
     In summary, :math:`s =  \sum_i \text{softmax}_j(a_i) n_i` is computed by the layer.
-
     """
 
     def __init__(self, **kwargs):
         """Initialize layer."""
-        super(PoolingNodesAttention, self).__init__(**kwargs)
+        super(PoolingEmbeddingAttention, self).__init__(**kwargs)
 
     def build(self, input_shape):
         """Build layer."""
-        super(PoolingNodesAttention, self).build(input_shape)
+        super(PoolingEmbeddingAttention, self).build(input_shape)
 
     def call(self, inputs, **kwargs):
         """Forward pass.
@@ -575,10 +577,11 @@ class PoolingNodesAttention(GraphBaseLayer):
         Returns:
             tf.Tensor: Embedding tensor of pooled node of shape (batch, F)
         """
-        dyn_inputs = inputs
+        # Need ragged input but can be generalized in the future.
+        self.assert_ragged_input_rank(inputs)
         # We cast to values here
-        nod, batchi, target_len = dyn_inputs[0].values, dyn_inputs[0].value_rowids(), dyn_inputs[0].row_lengths()
-        ats = dyn_inputs[1].values
+        nod, batchi, target_len = inputs[0].values, inputs[0].value_rowids(), inputs[0].row_lengths()
+        ats = inputs[1].values
 
         ats = segment_softmax(ats, batchi)
         get = nod * ats
@@ -588,5 +591,8 @@ class PoolingNodesAttention(GraphBaseLayer):
 
     def get_config(self):
         """Update layer config."""
-        config = super(PoolingNodesAttention, self).get_config()
+        config = super(PoolingEmbeddingAttention, self).get_config()
         return config
+
+
+PoolingNodesAttention = PoolingEmbeddingAttention
