@@ -4,9 +4,9 @@ import tensorflow.keras as ks
 from kgcnn.layers.casting import ChangeTensorType
 from kgcnn.layers.conv.attention import AttentiveHeadFP, PoolingNodesAttentive
 from kgcnn.layers.conv.mpnn_conv import GRUUpdate
-from kgcnn.layers.keras import Dense, Dropout
-from kgcnn.layers.mlp import MLP
-from kgcnn.utils.models import generate_embedding, update_model_kwargs
+from kgcnn.layers.modules import DenseEmbedding, DropoutEmbedding, OptionalInputEmbedding
+from kgcnn.layers.mlp import GraphMLP, MLP
+from kgcnn.utils.models import update_model_kwargs
 
 # Pushing the Boundaries of Molecular Representation for Drug Discovery with the Graph Attention Mechanism
 # Zhaoping Xiong, Dingyan Wang, Xiaohong Liu, Feisheng Zhong, Xiaozhe Wan, Xutong Li, Zhaojun Li,
@@ -21,37 +21,41 @@ model_default = {'name': "AttentiveFP",
                             {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                  'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                      "edge": {"input_dim": 5, "output_dim": 64}},
-                 'output_embedding': 'graph',
-                 'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
-                                "activation": ['relu', 'relu', 'sigmoid']},
                  'attention_args': {"units": 32},
                  'depth': 3,
                  'dropout': 0.1,
-                 'verbose': 1
+                 'verbose': 10,
+                 'output_embedding': 'graph',
+                 'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
+                                "activation": ['relu', 'relu', 'sigmoid']}
                  }
 
 
 @update_model_kwargs(model_default)
 def make_model(inputs=None,
+               input_embedding=None,
                depth=None,
                dropout=None,
-               input_embedding=None,
-               output_embedding=None,
-               output_mlp=None,
                attention_args=None,
-               **kwargs):
+               name=None,
+               verbose=None,
+               output_embedding=None,
+               output_mlp=None
+               ):
     """Make AttentiveFP graph network via functional API. Default parameters can be found in :obj:`model_default`.
 
     Args:
         inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
+        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
         depth (int): Number of graph embedding units or depth of the network.
         dropout (float): Dropout to use.
-        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
+        attention_args (dict): Dictionary of layer arguments unpacked in `AttentiveHeadFP` layer. Units parameter
+            is also used in GRU-update and `PoolingNodesAttentive`.
+        name (str): Name of the model.
+        verbose (int): Level of print output.
         output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
         output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
             Defines number of model outputs and activation.
-        attention_args (dict): Dictionary of layer arguments unpacked in `AttentiveHeadFP` layer. Units parameter
-            is also used in GRU-update and `PoolingNodesAttentive`.
 
     Returns:
         tf.keras.models.Model
@@ -63,28 +67,29 @@ def make_model(inputs=None,
     edge_index_input = ks.layers.Input(**inputs[2])
 
     # Embedding, if no feature dimension
-    n = generate_embedding(node_input, inputs[0]['shape'], input_embedding['node'])
-    ed = generate_embedding(edge_input, inputs[1]['shape'], input_embedding['edge'])
+    n = OptionalInputEmbedding(**input_embedding['node'],
+                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
+    ed = OptionalInputEmbedding(**input_embedding['edge'],
+                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
     edi = edge_index_input
 
     # Model
-    nk = Dense(units=attention_args['units'])(n)
+    nk = DenseEmbedding(units=attention_args['units'])(n)
     ck = AttentiveHeadFP(use_edge_features=True, **attention_args)([nk, ed, edi])
     nk = GRUUpdate(units=attention_args['units'])([nk, ck])
 
     for i in range(1, depth):
         ck = AttentiveHeadFP(**attention_args)([nk, ed, edi])
         nk = GRUUpdate(units=attention_args['units'])([nk, ck])
-        nk = Dropout(rate=dropout)(nk)
+        nk = DropoutEmbedding(rate=dropout)(nk)
     n = nk
 
     # Output embedding choice
     if output_embedding == 'graph':
         out = PoolingNodesAttentive(units=attention_args['units'])(n)
-        out = MLP(**output_mlp)(out)
-        main_output = ks.layers.Flatten()(out)  # will be dense
+        main_output = MLP(**output_mlp)(out)
     elif output_embedding == 'node':  # node embedding
-        out = MLP(**output_mlp)(n)
+        out = GraphMLP(**output_mlp)(n)
         main_output = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
     else:
         raise ValueError("Unsupported graph embedding for mode `AttentiveFP`")
