@@ -2,9 +2,9 @@ import tensorflow.keras as ks
 
 from kgcnn.layers.casting import ChangeTensorType
 from kgcnn.layers.conv.gcn_conv import GCN
-from kgcnn.layers.keras import Dense
-from kgcnn.layers.mlp import MLP
-from kgcnn.layers.pool.pooling import PoolingNodes, PoolingWeightedNodes
+from kgcnn.layers.modules import DenseEmbedding, OptionalInputEmbedding
+from kgcnn.layers.mlp import GraphMLP, MLP
+from kgcnn.layers.pooling import PoolingNodes, PoolingWeightedNodes
 from kgcnn.utils.models import update_model_kwargs, generate_embedding
 
 # 'Semi-Supervised Classification with Graph Convolutional Networks'
@@ -18,38 +18,41 @@ model_default = {'name': "GCN",
                             {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                  'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                      "edge": {"input_dim": 10, "output_dim": 64}},
-                 'output_embedding': 'graph',
-                 'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
-                                "activation": ['relu', 'relu', 'sigmoid']},
                  'gcn_args': {"units": 100, "use_bias": True, "activation": 'relu', "pooling_method": 'sum',
                               "is_sorted": False, "has_unconnected": True},
-                 'depth': 3, 'verbose': 1
+                 'depth': 3, 'verbose': 10,
+                 'output_embedding': 'graph',
+                 'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
+                                "activation": ['relu', 'relu', 'sigmoid']}
                  }
 
 
 @update_model_kwargs(model_default)
 def make_model(inputs=None,
                input_embedding=None,
-               output_embedding=None,
-               output_mlp=None,
                depth=None,
                gcn_args=None,
-               **kwargs):
+               name=None,
+               verbose=None,
+               output_embedding=None,
+               output_mlp=None
+               ):
     """Make GCN graph network via functional API. Default parameters can be found in :obj:`model_default`.
 
     Args:
         inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
         input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
+        depth (int): Number of graph embedding units or depth of the network.
+        gcn_args (dict): Dictionary of layer arguments unpacked in `GCN` convolutional layer.
+        name (str): Name of the model.
+        verbose (int): Level of print output.
         output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
         output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
             Defines number of model outputs and activation.
-        depth (int): Number of graph embedding units or depth of the network.
-        gcn_args (dict): Dictionary of layer arguments unpacked in `GCN` convolutional layer.
 
     Returns:
         tf.keras.models.Model
     """
-
     input_node_shape = inputs[0]['shape']
     input_edge_shape = inputs[1]['shape']
 
@@ -63,12 +66,14 @@ def make_model(inputs=None,
     edge_index_input = ks.layers.Input(**inputs[2])
 
     # Embedding, if no feature dimension
-    n = generate_embedding(node_input, input_node_shape, input_embedding['node'])
-    ed = generate_embedding(edge_input, input_edge_shape, input_embedding['edge'])
+    n = OptionalInputEmbedding(**input_embedding['node'],
+                               use_embedding=len(input_node_shape) < 2)(node_input)
+    ed = OptionalInputEmbedding(**input_embedding['edge'],
+                                use_embedding=len(input_edge_shape) < 2)(edge_input)
     edi = edge_index_input
 
     # Model
-    n = Dense(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
+    n = DenseEmbedding(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
     for i in range(0, depth):
         n = GCN(**gcn_args)([n, ed, edi])
 
@@ -78,9 +83,9 @@ def make_model(inputs=None,
         out = MLP(**output_mlp)(out)
     elif output_embedding == "node":
         out = n
-        out = MLP(**output_mlp)(out)
-        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(
-            out)  # no ragged for distribution supported atm
+        out = GraphMLP(**output_mlp)(out)
+        # no ragged for distribution supported atm
+        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
     else:
         raise ValueError("Unsupported graph embedding for `GCN`")
 
@@ -110,8 +115,9 @@ def make_model_weighted(inputs=None,
                         output_mlp=None,
                         depth=None,
                         gcn_args=None,
-                        **kwargs):
-    """Make GCN model."""
+                        name=None,
+                        verbose=None):
+    """Make GCN model with edge weights."""
 
     input_node_shape = inputs[0]['shape']
     input_edge_shape = inputs[1]['shape']
@@ -133,7 +139,7 @@ def make_model_weighted(inputs=None,
     nw = node_weights_input
 
     # Model
-    n = Dense(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
+    n = DenseEmbedding(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
     for i in range(0, depth):
         n = GCN(**gcn_args)([n, ed, edi])
 
@@ -143,9 +149,9 @@ def make_model_weighted(inputs=None,
         out = MLP(**output_mlp)(out)
     elif output_embedding == "node":
         out = n
-        out = MLP(**output_mlp)(out)
-        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(
-            out)  # no ragged for distribution supported atm
+        out = GraphMLP(**output_mlp)(out)
+        # no ragged for distribution supported atm
+        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
     else:
         raise ValueError("Unsupported graph embedding for `GCN`")
 
