@@ -2,11 +2,11 @@ import tensorflow.keras as ks
 
 from kgcnn.layers.casting import ChangeTensorType
 from kgcnn.layers.gather import GatherState, GatherNodesIngoing, GatherNodesOutgoing
-from kgcnn.layers.keras import Concatenate, Dense
-from kgcnn.layers.mlp import MLP
-from kgcnn.layers.pool.pooling import PoolingLocalEdges, PoolingNodes
+from kgcnn.layers.modules import LazyConcatenate, DenseEmbedding, OptionalInputEmbedding
+from kgcnn.layers.mlp import GraphMLP, MLP
+from kgcnn.layers.pooling import PoolingLocalEdges, PoolingNodes
 from kgcnn.layers.pool.set2set import PoolingSet2Set
-from kgcnn.utils.models import update_model_kwargs, generate_embedding
+from kgcnn.utils.models import update_model_kwargs
 
 # 'Interaction Networks for Learning about Objects,Relations and Physics'
 # by Peter W. Battaglia, Razvan Pascanu, Matthew Lai, Danilo Rezende, Koray Kavukcuoglu
@@ -21,25 +21,23 @@ hyper_model_default = {'name': "INorp",
                        'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                            "edge": {"input_dim": 5, "output_dim": 64},
                                            "graph": {"input_dim": 100, "output_dim": 64}},
-                       'output_embedding': 'graph',
-                       'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
-                                      "activation": ['relu', 'relu', 'sigmoid']},
-                       'set2set_args': {'channels': 32, 'T': 3, "pooling_method": "mean",
+                       'set2set_args': {"channels": 32, "T": 3, "pooling_method": "mean",
                                         "init_qstar": "mean"},
                        'node_mlp_args': {"units": [100, 50], "use_bias": True, "activation": ['relu', "linear"]},
                        'edge_mlp_args': {"units": [100, 100, 100, 100, 50],
                                          "activation": ['relu', 'relu', 'relu', 'relu', "linear"]},
                        'pooling_args': {'pooling_method': "segment_mean"},
-                       'depth': 3, 'use_set2set': False, 'verbose': 1,
-                       'gather_args': {}
+                       'depth': 3, 'use_set2set': False, 'verbose': 10,
+                       'gather_args': {},
+                       'output_embedding': 'graph',
+                       'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
+                                      "activation": ['relu', 'relu', 'sigmoid']}
                        }
 
 
 @update_model_kwargs(hyper_model_default)
 def make_model(inputs=None,
                input_embedding=None,
-               output_embedding=None,
-               output_mlp=None,
                depth=None,
                gather_args=None,
                edge_mlp_args=None,
@@ -47,16 +45,16 @@ def make_model(inputs=None,
                set2set_args=None,
                pooling_args=None,
                use_set2set=None,
-               **kwargs
+               name=None,
+               verbose=None,
+               output_embedding=None,
+               output_mlp=None
                ):
     """Make INorp graph network via functional API. Default parameters can be found in :obj:`model_default`.
 
     Args:
         inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
         input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
-        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
-        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
-            Defines number of model outputs and activation.
         depth (int): Number of graph embedding units or depth of the network.
         gather_args (dict): Dictionary of layer arguments unpacked in `GatherNodes` layer.
         edge_mlp_args (dict): Dictionary of layer arguments unpacked in `MLP` layer for edge updates.
@@ -64,6 +62,11 @@ def make_model(inputs=None,
         set2set_args (dict): Dictionary of layer arguments unpacked in `PoolingSet2Set` layer.
         pooling_args (dict): Dictionary of layer arguments unpacked in `PoolingLocalEdges`, `PoolingNodes` layer.
         use_set2set (bool): Whether to use `PoolingSet2Set` layer.
+        verbose (int): Level of verbosity.
+        name (str): Name of the model.
+        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
+        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
+            Defines number of model outputs and activation.
 
     Returns:
         tf.keras.models.Model
@@ -76,9 +79,12 @@ def make_model(inputs=None,
     env_input = ks.Input(**inputs[3])
 
     # embedding, if no feature dimension
-    n = generate_embedding(node_input, inputs[0]['shape'], input_embedding['node'])
-    ed = generate_embedding(edge_input, inputs[1]['shape'], input_embedding['edge'])
-    uenv = generate_embedding(env_input, inputs[3]['shape'], input_embedding['graph'], embedding_rank=0)
+    n = OptionalInputEmbedding(**input_embedding['node'],
+                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
+    ed = OptionalInputEmbedding(**input_embedding['edge'],
+                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
+    uenv = OptionalInputEmbedding(**input_embedding['graph'],
+                                  use_embedding=len(inputs[3]['shape']) < 1)(env_input)
     edi = edge_index_input
 
     # Model
@@ -88,30 +94,31 @@ def make_model(inputs=None,
         # upd = GatherNodes()([n,edi])
         eu1 = GatherNodesIngoing(**gather_args)([n, edi])
         eu2 = GatherNodesOutgoing(**gather_args)([n, edi])
-        upd = Concatenate(axis=-1)([eu2, eu1])
-        eu = Concatenate(axis=-1)([upd, ed])
+        upd = LazyConcatenate(axis=-1)([eu2, eu1])
+        eu = LazyConcatenate(axis=-1)([upd, ed])
 
-        eu = MLP(**edge_mlp_args)(eu)
+        eu = GraphMLP(**edge_mlp_args)(eu)
         # Pool message
         nu = PoolingLocalEdges(**pooling_args)(
             [n, eu, edi])  # Summing for each node connection
         # Add environment
-        nu = Concatenate(axis=-1)(
-            [n, nu, ev])  # Concatenate node features with new edge updates
-        n = MLP(**node_mlp_args)(nu)
+        nu = LazyConcatenate(axis=-1)(
+            [n, nu, ev])  # LazyConcatenate node features with new edge updates
+        n = GraphMLP(**node_mlp_args)(nu)
 
     # Output embedding choice
     if output_embedding == 'graph':
         if use_set2set:
             # output
-            outss = Dense(set2set_args["channels"], activation="linear")(n)
-            out = PoolingSet2Set(**set2set_args)(outss)
+            n = DenseEmbedding(set2set_args["channels"], activation="linear")(n)
+            out = PoolingSet2Set(**set2set_args)(n)
         else:
             out = PoolingNodes(**pooling_args)(n)
+        out = ks.layers.Flatten()(out)
         main_output = MLP(**output_mlp)(out)
     elif output_embedding == 'node':
         out = n
-        main_output = MLP(**output_mlp)(out)
+        main_output = GraphMLP(**output_mlp)(out)
         main_output = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(main_output)
         # no ragged for distribution atm
     else:
