@@ -1,12 +1,12 @@
 import tensorflow.keras as ks
 
 from kgcnn.layers.conv.megnet_conv import MEGnetBlock
-from kgcnn.layers.geom import NodeDistance, GaussBasisLayer
-from kgcnn.layers.keras import Dense, Add, Dropout
-from kgcnn.layers.mlp import MLP
-from kgcnn.layers.pool.pooling import PoolingGlobalEdges, PoolingNodes
+from kgcnn.layers.geom import NodeDistanceEuclidean, GaussBasisLayer, NodePosition
+from kgcnn.layers.modules import DenseEmbedding, LazyAdd, DropoutEmbedding, OptionalInputEmbedding
+from kgcnn.layers.mlp import GraphMLP, MLP
+from kgcnn.layers.pooling import PoolingGlobalEdges, PoolingNodes
 from kgcnn.layers.pool.set2set import PoolingSet2Set
-from kgcnn.utils.models import generate_embedding, update_model_kwargs
+from kgcnn.utils.models import update_model_kwargs
 
 # from kgcnn.layers.casting import ChangeTensorType, ChangeIndexing
 
@@ -22,27 +22,27 @@ model_default = {'name': "Megnet",
                             {'shape': [], 'name': "graph_attributes", 'dtype': 'float32', 'ragged': False}],
                  'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                      "graph": {"input_dim": 100, "output_dim": 64}},
-                 'output_embedding': 'graph',
-                 'output_mlp': {"use_bias": [True, True, True], "units": [32, 16, 1],
-                                "activation": ['kgcnn>softplus2', 'kgcnn>softplus2', 'linear']},
+                 "make_distance": True, "expand_distance": True,
                  'gauss_args': {"bins": 20, "distance": 4, "offset": 0.0, "sigma": 0.4},
                  'meg_block_args': {'node_embed': [64, 32, 32], 'edge_embed': [64, 32, 32],
                                     'env_embed': [64, 32, 32], 'activation': 'kgcnn>softplus2'},
                  'set2set_args': {'channels': 16, 'T': 3, "pooling_method": "sum", "init_qstar": "0"},
                  'node_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
                  'edge_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
-                 'state_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2",
-                                   "input_tensor_type": "tensor"},
+                 'state_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
                  'nblocks': 3, 'has_ff': True, 'dropout': None, 'use_set2set': True,
-                 'verbose': 1
+                 'verbose': 10,
+                 'output_embedding': 'graph',
+                 'output_mlp': {"use_bias": [True, True, True], "units": [32, 16, 1],
+                                "activation": ['kgcnn>softplus2', 'kgcnn>softplus2', 'linear']}
                  }
 
 
 @update_model_kwargs(model_default)
 def make_model(inputs=None,
                input_embedding=None,
-               output_embedding=None,
-               output_mlp=None,
+               expand_distance=None,
+               make_distance=None,
                gauss_args=None,
                meg_block_args=None,
                set2set_args=None,
@@ -53,16 +53,19 @@ def make_model(inputs=None,
                nblocks=None,
                has_ff=None,
                dropout=None,
-               **kwargs
+               name=None,
+               verbose=None,
+               output_embedding=None,
+               output_mlp=None
                ):
-    """Make Megnet graph network via functional API. Default parameters can be found in :obj:`model_default`.
+    """Make MegNet graph network via functional API. Default parameters can be found in :obj:`model_default`.
 
     Args:
         inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
         input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
-        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
-        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
-            Defines number of model outputs and activation.
+        make_distance (bool): Whether input is distance or coordinates at in place of edges.
+        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
+            form edges with a gauss distance basis given edge indices indices. Expansion uses `gauss_args`.
         gauss_args (dict): Dictionary of layer arguments unpacked in `GaussBasisLayer` layer.
         meg_block_args (dict): Dictionary of layer arguments unpacked in `MEGnetBlock` layer.
         set2set_args (dict): Dictionary of layer arguments unpacked in `PoolingSet2Set` layer.
@@ -73,6 +76,11 @@ def make_model(inputs=None,
         nblocks (int): Number of graph embedding blocks or depth of the network.
         has_ff (bool): Use feed-forward MLP in each block.
         dropout (int): Dropout to use. Default is None.
+        name (str): Name of the model.
+        verbose (int): Verbosity level of print.
+        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
+        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
+            Defines number of model outputs and activation.
 
     Returns:
         tf.keras.models.Model
@@ -85,29 +93,37 @@ def make_model(inputs=None,
     env_input = ks.Input(**inputs[3])
 
     # embedding, if no feature dimension
-    n = generate_embedding(node_input, inputs[0]['shape'], input_embedding['node'])
-    uenv = generate_embedding(env_input, inputs[3]['shape'], input_embedding['graph'], embedding_rank=0)
+    n = OptionalInputEmbedding(**input_embedding['node'],
+                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
+    uenv = OptionalInputEmbedding(**input_embedding['graph'],
+                                  use_embedding=len(inputs[3]['shape']) < 1)(env_input)
     edi = edge_index_input
 
     # Edge distance as Gauss-Basis
-    x = xyz_input
-    ed = NodeDistance()([x, edi])
-    ed = GaussBasisLayer(**gauss_args)(ed)
+    if make_distance:
+        x = xyz_input
+        pos1, pos2 = NodePosition()([x, edi])
+        ed = NodeDistanceEuclidean()([pos1, pos2])
+    else:
+        ed = xyz_input
+
+    if expand_distance:
+        ed = GaussBasisLayer(**gauss_args)(ed)
 
     # Model
     vp = n
     ep = ed
     up = uenv
-    vp = MLP(**node_ff_args)(vp)
-    ep = MLP(**edge_ff_args)(ep)
+    vp = GraphMLP(**node_ff_args)(vp)
+    ep = GraphMLP(**edge_ff_args)(ep)
     up = MLP(**state_ff_args)(up)
     vp2 = vp
     ep2 = ep
     up2 = up
     for i in range(0, nblocks):
         if has_ff and i > 0:
-            vp2 = MLP(**node_ff_args)(vp)
-            ep2 = MLP(**edge_ff_args)(ep)
+            vp2 = GraphMLP(**node_ff_args)(vp)
+            ep2 = GraphMLP(**edge_ff_args)(ep)
             up2 = MLP(**state_ff_args)(up)
 
         # MEGnetBlock
@@ -116,17 +132,17 @@ def make_model(inputs=None,
 
         # skip connection
         if dropout is not None:
-            vp2 = Dropout(dropout, name='dropout_atom_%d' % i)(vp2)
-            ep2 = Dropout(dropout, name='dropout_bond_%d' % i)(ep2)
-            up2 = Dropout(dropout, name='dropout_state_%d' % i)(up2)
+            vp2 = DropoutEmbedding(dropout, name='dropout_atom_%d' % i)(vp2)
+            ep2 = DropoutEmbedding(dropout, name='dropout_bond_%d' % i)(ep2)
+            up2 = ks.layers.Dropout(dropout, name='dropout_state_%d' % i)(up2)
 
-        vp = Add()([vp2, vp])
-        ep = Add()([ep2, ep])
-        up = Add(input_tensor_type="tensor")([up2, up])
+        vp = LazyAdd()([vp2, vp])
+        ep = LazyAdd()([ep2, ep])
+        up = ks.layers.Add()([up2, up])
 
     if use_set2set:
-        vp = Dense(set2set_args["channels"], activation='linear')(vp)  # to match units
-        ep = Dense(set2set_args["channels"], activation='linear')(ep)  # to match units
+        vp = DenseEmbedding(set2set_args["channels"], activation='linear')(vp)  # to match units
+        ep = DenseEmbedding(set2set_args["channels"], activation='linear')(ep)  # to match units
         vp = PoolingSet2Set(**set2set_args)(vp)
         ep = PoolingSet2Set(**set2set_args)(ep)
     else:
@@ -140,10 +156,10 @@ def make_model(inputs=None,
     if dropout is not None:
         final_vec = ks.layers.Dropout(dropout, name='dropout_final')(final_vec)
 
+    # Only graph embedding for Megnet
     if output_embedding != "graph":
         raise ValueError("Unsupported graph embedding for mode `Megnet`.")
-    # final dense layers
-    # Only graph embedding for Megnet
-    main_output = MLP(**output_mlp, input_tensor_type="tensor")(final_vec)
+
+    main_output = MLP(**output_mlp)(final_vec)
     model = ks.models.Model(inputs=[node_input, xyz_input, edge_index_input, env_input], outputs=main_output)
     return model
