@@ -1,13 +1,12 @@
 import tensorflow.keras as ks
 
 from kgcnn.layers.casting import ChangeTensorType
-from kgcnn.layers.connect import AdjacencyPower
 from kgcnn.layers.gather import GatherNodesOutgoing
-from kgcnn.layers.keras import Dense, Activation, Add
-from kgcnn.layers.mlp import MLP
-from kgcnn.layers.pool.pooling import PoolingNodes, PoolingLocalEdges
-from kgcnn.layers.pool.topk import PoolingTopK, UnPoolingTopK
-from kgcnn.utils.models import generate_embedding, update_model_kwargs
+from kgcnn.layers.modules import DenseEmbedding, ActivationEmbedding, LazyAdd, OptionalInputEmbedding
+from kgcnn.layers.mlp import GraphMLP, MLP
+from kgcnn.layers.pooling import PoolingNodes, PoolingLocalEdges
+from kgcnn.layers.pool.topk import PoolingTopK, UnPoolingTopK, AdjacencyPower
+from kgcnn.utils.models import update_model_kwargs
 
 # Graph U-Nets
 # by Hongyang Gao, Shuiwang Ji
@@ -19,8 +18,6 @@ model_default = {'name': "Unet",
                             {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                  'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                      "edge": {"input_dim": 5, "output_dim": 64}},
-                 'output_embedding': 'graph',
-                 'output_mlp': {"use_bias": [True, False], "units": [25, 1], "activation": ['relu', 'sigmoid']},
                  'hidden_dim': {'units': 32, 'use_bias': True, 'activation': 'linear'},
                  'top_k_args': {'k': 0.3, 'kernel_initializer': 'ones'},
                  'activation': 'relu',
@@ -28,30 +25,32 @@ model_default = {'name': "Unet",
                  'depth': 4,
                  'pooling_args': {"pooling_method": 'segment_mean'},
                  'gather_args': {"node_indexing": 'sample'},
-                 'verbose': 1
+                 'verbose': 10,
+                 'output_embedding': 'graph',
+                 'output_mlp': {"use_bias": [True, False], "units": [25, 1], "activation": ['relu', 'sigmoid']}
                  }
 
 
 @update_model_kwargs(model_default)
 def make_model(inputs=None,
                input_embedding=None,
-               output_embedding=None,
-               output_mlp=None,
                pooling_args=None,
                gather_args=None,
                top_k_args=None,
                depth=None,
                use_reconnect=None,
                hidden_dim=None,
-               activation=None, **kwargs):
+               activation=None,
+               name=None,
+               verbose=None,
+               output_embedding=None,
+               output_mlp=None
+               ):
     r"""Make U-Net graph network via functional API. Default parameters can be found in :obj:`model_default`.
 
     Args:
         inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
         input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
-        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
-        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
-            Defines number of model outputs and activation.
         depth (int): Number of graph embedding units or depth of the network.
         pooling_args (dict): Dictionary of layer arguments unpacked in `PoolingLocalEdges` layers.
         gather_args (dict): Dictionary of layer arguments unpacked in `GatherNodesOutgoing` layers.
@@ -59,6 +58,11 @@ def make_model(inputs=None,
         use_reconnect (bool): Whether to use :math:`A^2` between pooling.
         hidden_dim (dict): Dictionary of layer arguments unpacked in hidden `Dense` layer.
         activation (dict, str): Activation to use.
+        verbose (int): Level of verbosity.
+        name (str): Name of the model.
+        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
+        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
+            Defines number of model outputs and activation.
 
     Returns:
         tf.keras.models.Model
@@ -70,12 +74,14 @@ def make_model(inputs=None,
     edge_index_input = ks.layers.Input(**inputs[2])
 
     # embedding, if no feature dimension
-    n = generate_embedding(node_input, inputs[0]['shape'], input_embedding['node'])
-    ed = generate_embedding(edge_input, inputs[1]['shape'], input_embedding['edge'])
+    n = OptionalInputEmbedding(**input_embedding['node'],
+                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
+    ed = OptionalInputEmbedding(**input_embedding['edge'],
+                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
     edi = edge_index_input
 
     # Model
-    n = Dense(**hidden_dim)(n)
+    n = DenseEmbedding(**hidden_dim)(n)
     in_graph = [n, ed, edi]
     graph_list = [in_graph]
     map_list = []
@@ -87,9 +93,9 @@ def make_model(inputs=None,
         n, ed, edi = i_graph
         # GCN layer
         eu = GatherNodesOutgoing(**gather_args)([n, edi])
-        eu = Dense(**hidden_dim)(eu)
+        eu = DenseEmbedding(**hidden_dim)(eu)
         nu = PoolingLocalEdges(**pooling_args)([n, eu, edi])  # Summing for each node connection
-        n = Activation(activation=activation)(nu)
+        n = ActivationEmbedding(activation=activation)(nu)
 
         if use_reconnect:
             ed, edi = AdjacencyPower(n=2)([n, ed, edi])
@@ -109,12 +115,12 @@ def make_model(inputs=None,
 
         n, ed, edi = ui_graph
         # skip connection
-        n = Add()([n, o_graph[0]])
+        n = LazyAdd()([n, o_graph[0]])
         # GCN
         eu = GatherNodesOutgoing(**gather_args)([n, edi])
-        eu = Dense(**hidden_dim)(eu)
+        eu = DenseEmbedding(**hidden_dim)(eu)
         nu = PoolingLocalEdges(**pooling_args)([n, eu, edi])  # Summing for each node connection
-        n = Activation(activation=activation)(nu)
+        n = ActivationEmbedding(activation=activation)(nu)
 
         ui_graph = [n, ed, edi]
 
@@ -122,10 +128,10 @@ def make_model(inputs=None,
     n = ui_graph[0]
     if output_embedding == 'graph':
         out = PoolingNodes(**pooling_args)(n)
-        out = MLP(**output_mlp)(out)
-        main_output = ks.layers.Flatten()(out)  # will be dense
+        out = ks.layers.Flatten()(out)  # will be dense
+        main_output = MLP(**output_mlp)(out)
     elif output_embedding == 'node':
-        out = MLP(**output_mlp)(n)
+        out = GraphMLP(**output_mlp)(n)
         main_output = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
     else:
         raise ValueError("Unsupported graph embedding for mode `Unet`")
