@@ -2,6 +2,22 @@ import numpy as np
 from sklearn.linear_model import Ridge
 import matplotlib.pyplot as plt
 
+global_proton_dict = {'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10, 'Na': 11,
+                      'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
+                      'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29,
+                      'Zn': 30, 'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38,
+                      'Y': 39, 'Zr': 40, 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47,
+                      'Cd': 48, 'In': 49, 'Sn': 50, 'Sb': 51, 'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56,
+                      'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60, 'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65,
+                      'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70, 'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74,
+                      'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83,
+                      'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90, 'Pa': 91, 'U': 92,
+                      'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100, 'Md': 101,
+                      'No': 102, 'Lr': 103, 'Rf': 104, 'Db': 105, 'Sg': 106, 'Bh': 107, 'Hs': 108, 'Mt': 109,
+                      'Ds': 110, 'Rg': 111, 'Cn': 112, 'Nh': 113, 'Fl': 114, 'Mc': 115, 'Lv': 116, 'Ts': 117,
+                      'Og': 118, 'Uue': 119}
+inverse_global_proton_dict = {value: key for key, value in global_proton_dict.items()}
+
 
 def get_connectivity_from_inverse_distance_matrix(inv_dist_mat, protons, radii_dict=None, k1=16.0, k2=4.0 / 3.0,
                                                   cutoff=0.85, force_bonds=True):
@@ -40,7 +56,7 @@ def get_connectivity_from_inverse_distance_matrix(inv_dist_mat, protons, radii_d
     if radii_dict is None:
         radii_dict = proton_radii_dict  # index matches atom number
     # Get Radii
-    protons = np.array(protons, dtype=np.int)
+    protons = np.array(protons, dtype="int")
     radii = radii_dict[protons]
     # Calculate
     shape_rad = radii.shape
@@ -94,7 +110,7 @@ class ExtensiveMolecularScaler:
         """Fit atomic number to the molecular properties.
 
         Args:
-            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, <#atoms>)`.
+            atomic_number (list): List of array of atomic numbers. Shape is `(n_samples, #atoms)`.
             molecular_property (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
             sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.
 
@@ -103,7 +119,7 @@ class ExtensiveMolecularScaler:
         """
         if len(atomic_number) != len(molecular_property):
             raise ValueError(
-                "ERROR:kgcnn: `ExtensiveMolecularScaler` different input shape {0} vs. {1}".format(
+                "`ExtensiveMolecularScaler` different input shape {0} vs. {1}".format(
                     len(atomic_number), len(molecular_property))
             )
 
@@ -123,7 +139,7 @@ class ExtensiveMolecularScaler:
         self.ridge.fit(total_number, molecular_property, sample_weight=sample_weight)
         self._fit_coef = self.ridge.coef_
         self._fit_intercept = self.ridge.intercept_
-        diff = molecular_property-self.ridge.predict(total_number)
+        diff = molecular_property - self.ridge.predict(total_number)
         self.scale_ = np.std(diff, axis=0, keepdims=True)
         return self
 
@@ -145,24 +161,25 @@ class ExtensiveMolecularScaler:
             array_atoms[unique_per_mol] = num_unique
             positives = array_atoms[self._fit_atom_selection_mask]
             if np.sum(positives) != np.sum(num_unique):
-                print("WARNING:kgcnn: `ExtensiveMolecularScaler` got unknown atom species in transform.")
+                print("`ExtensiveMolecularScaler` got unknown atom species in transform.")
             total_number.append(positives)
         total_number = np.array(total_number)
         offset = self.ridge.predict(total_number)
         return offset
 
-    def _plot_predict(self, atomic_number,  molecular_property):
+    def _plot_predict(self, atomic_number, molecular_property):
         """Debug function to check prediction."""
+        molecular_property = np.array(molecular_property)
         if len(molecular_property.shape) <= 1:
             molecular_property = np.expand_dims(molecular_property, axis=-1)
         predict_prop = self.predict(atomic_number)
         if len(predict_prop.shape) <= 1:
             predict_prop = np.expand_dims(predict_prop, axis=-1)
-        mae = np.mean(np.abs(molecular_property-predict_prop), axis=0)
+        mae = np.mean(np.abs(molecular_property - predict_prop), axis=0)
         plt.figure()
         for i in range(predict_prop.shape[-1]):
             plt.scatter(predict_prop[:, i], molecular_property[:, i], alpha=0.3,
-                        label="Pos: "+str(i) + " MAE: {0:0.4f} ".format(mae[i]))
+                        label="Pos: " + str(i) + " MAE: {0:0.4f} ".format(mae[i]))
         plt.plot(np.arange(np.amin(molecular_property), np.amax(molecular_property), 0.05),
                  np.arange(np.amin(molecular_property), np.amax(molecular_property), 0.05), color='red')
         plt.xlabel('Fitted')
@@ -180,7 +197,7 @@ class ExtensiveMolecularScaler:
         Returns:
             np.ndarray: Transformed atomic properties fitted. Shape is `(n_samples, n_properties)`.
         """
-        return (molecular_property-self.predict(atomic_number))/self.scale_
+        return (molecular_property - self.predict(atomic_number)) / self.scale_
 
     def fit_transform(self, atomic_number, molecular_property, sample_weight=None):
         """Combine fit and transform methods in one call.
@@ -206,4 +223,4 @@ class ExtensiveMolecularScaler:
         Returns:
             np.ndarray: Original atomic properties. Shape is `(n_samples, n_properties)`.
         """
-        return molecular_property*self.scale_+self.predict(atomic_number)
+        return molecular_property * self.scale_ + self.predict(atomic_number)
