@@ -5,7 +5,7 @@ import scipy.sparse as sp
 def precompute_adjacency_scaled(adj_matrix, add_identity: bool = True):
     r"""Precompute the scaled adjacency matrix :math:`A_{s} = D^{-0.5} (A + I) D^{-0.5}`
     after Thomas N. Kipf and Max Welling (2016). Where :math:`I` denotes the diagonal unity matrix.
-    The node degree matrix is defined as :math:`D_{ii} = \sum_{j} (A + I)_{ij}`.
+    The node degree matrix is defined as :math:`D_{i,i} = \sum_{j} (A + I)_{i,j}`.
 
     Args:
         adj_matrix (np.ndarray, scipy.sparse): Adjacency matrix :math:`A` of shape `(N, N)`.
@@ -15,15 +15,16 @@ def precompute_adjacency_scaled(adj_matrix, add_identity: bool = True):
         array-like: Scaled adjacency matrix after :math:`A_{s} = D^{-0.5} (A + I) D^{-0.5}`.
     """
     if isinstance(adj_matrix, np.ndarray):
-        adj_matrix = np.array(adj_matrix, dtype=np.float)
+        adj_matrix = np.array(adj_matrix, dtype="float")
         if add_identity:
             adj_matrix = adj_matrix + np.identity(adj_matrix.shape[0])
         rowsum = np.sum(adj_matrix, axis=-1)
         colsum = np.sum(adj_matrix, axis=0)
-        d_ii = np.power(rowsum, -0.5).flatten()
-        d_jj = np.power(colsum, -0.5).flatten()
-        d_ii[np.isinf(d_ii)] = 0.
-        d_jj[np.isinf(d_jj)] = 0.
+        with np.errstate(divide='ignore', invalid='ignore'):
+            d_ii = np.power(rowsum, -0.5).flatten()
+            d_jj = np.power(colsum, -0.5).flatten()
+            d_ii = np.nan_to_num(d_ii, nan=0.0, posinf=0.0, neginf=0.0)
+            d_jj = np.nan_to_num(d_jj, nan=0.0, posinf=0.0, neginf=0.0)
         di = np.zeros((adj_matrix.shape[0], adj_matrix.shape[0]), dtype=adj_matrix.dtype)
         dj = np.zeros((adj_matrix.shape[1], adj_matrix.shape[1]), dtype=adj_matrix.dtype)
         di[np.arange(adj_matrix.shape[0]), np.arange(adj_matrix.shape[0])] = d_ii
@@ -35,10 +36,11 @@ def precompute_adjacency_scaled(adj_matrix, add_identity: bool = True):
             adj = adj + sp.eye(adj.shape[0])
         colsum = np.array(adj.sum(0))
         rowsum = np.array(adj.sum(1))
-        d_ii = np.power(rowsum, -0.5).flatten()
-        d_jj = np.power(colsum, -0.5).flatten()
-        d_ii[np.isinf(d_ii)] = 0.
-        d_jj[np.isinf(d_jj)] = 0.
+        with np.errstate(divide='ignore', invalid='ignore'):
+            d_ii = np.power(rowsum, -0.5).flatten()
+            d_jj = np.power(colsum, -0.5).flatten()
+            d_ii = np.nan_to_num(d_ii, nan=0.0, posinf=0.0, neginf=0.0)
+            d_jj = np.nan_to_num(d_jj, nan=0.0, posinf=0.0, neginf=0.0)
         di = sp.diags(d_ii, format='coo')
         dj = sp.diags(d_jj, format='coo')
         return di.dot(adj).dot(dj).tocoo()
@@ -46,6 +48,36 @@ def precompute_adjacency_scaled(adj_matrix, add_identity: bool = True):
         raise TypeError("Matrix format not supported: %s" % type(adj_matrix))
 
 
+def rescale_edge_weights_degree_sym(edge_indices, edge_weights):
+    r"""Normalize edge weights as :math:`\tilde(e)_{i,j} = d_{i,i}^{-0.5} e_{i,j} d_{j,j}^{-0.5}`.
+    The node degree is defined as :math:`D_{i,i} = \sum_{j} A_{i, j}`.
+
+
+    Args:
+        edge_indices (np.ndarray): Index-list referring to nodes of shape `(N, 2)`
+        edge_weights (np.ndarray): Edge weights matching indices of shape `(N, 1)`
+
+    Returns:
+        edge_weights (np.ndarray):  Rescaled edge weights of shape
+    """
+    if len(edge_indices) == 0:
+        return np.array([])
+    row_val, row_cnt = np.unique(edge_indices[:, 0], return_counts=True)
+    col_val, col_cnt = np.unique(edge_indices[:, 1], return_counts=True)
+    d_row = np.zeros(len(edge_weights), dtype=edge_weights.dtype)
+    d_col = np.zeros(len(edge_weights), dtype=edge_weights.dtype)
+    d_row[row_val] = row_cnt
+    d_col[col_val] = col_cnt
+    with np.errstate(divide='ignore', invalid='ignore'):
+        d_ii = np.power(d_row, -0.5).flatten()
+        d_jj = np.power(d_col, -0.5).flatten()
+        d_ii = np.nan_to_num(d_ii, nan=0.0, posinf=0.0, neginf=0.0)
+        d_jj = np.nan_to_num(d_jj, nan=0.0, posinf=0.0, neginf=0.0)
+    new_weights = np.expand_dims(d_ii[edge_indices[:, 0]], axis=-1) * edge_weights * np.expand_dims(
+        d_jj[edge_indices[:, 1]], axis=-1)
+    return new_weights
+
+
 def convert_scaled_adjacency_to_list(adj_scaled):
     r"""Map adjacency matrix to index list plus edge weights. In case of a standard adjacency matrix the edge weights
     will be one. For a pre-scaled adjacency matrix they become the entries of :math:`A_{s}`.
@@ -60,7 +92,7 @@ def convert_scaled_adjacency_to_list(adj_scaled):
             - edge_weight (np.ndarray): Entries of Adjacency matrix of shape `(N, )`.
     """
     if isinstance(adj_scaled, np.ndarray):
-        a = np.array(adj_scaled > 0, dtype=np.bool)
+        a = np.array(adj_scaled > 0, dtype="bool")
         edge_weight = adj_scaled[a]
         index1 = np.tile(np.expand_dims(np.arange(0, a.shape[0]), axis=1), (1, a.shape[1]))
         index2 = np.tile(np.expand_dims(np.arange(0, a.shape[1]), axis=0), (a.shape[0], 1))
@@ -69,8 +101,8 @@ def convert_scaled_adjacency_to_list(adj_scaled):
         return edge_index, edge_weight
     elif isinstance(adj_scaled, (sp.bsr.bsr_matrix, sp.csc.csc_matrix, sp.coo.coo_matrix, sp.csr.csr_matrix)):
         adj_scaled = adj_scaled.tocoo()
-        ei1 = np.array(adj_scaled.row.tolist(), dtype=np.int)
-        ei2 = np.array(adj_scaled.col.tolist(), dtype=np.int)
+        ei1 = np.array(adj_scaled.row.tolist(), dtype="int")
+        ei2 = np.array(adj_scaled.col.tolist(), dtype="int")
         edge_index = np.concatenate([np.expand_dims(ei1, axis=-1), np.expand_dims(ei2, axis=-1)], axis=-1)
         edge_weight = np.array(adj_scaled.data)
         return edge_index, edge_weight
@@ -101,8 +133,9 @@ def make_adjacency_undirected_logical_or(adj_mat):
         return a_out.tocoo()
 
 
-def add_self_loops_to_edge_indices(edge_indices, *args, remove_duplicates: bool = True, sort_indices: bool = True):
-    r"""Add self-loops to edge index list, i.e. `[0, 0], [1, 1], ...]`. Edge values are filled up with ones.
+def add_self_loops_to_edge_indices(edge_indices, *args, remove_duplicates: bool = True, sort_indices: bool = True,
+                                   fill_value: int = 0):
+    r"""Add self-loops to edge index list, i.e. `[0, 0], [1, 1], ...]`. Edge values are filled up with ones or zeros.
     Default mode is to remove duplicates in the added list. Edge indices are sorted by default. Sorting is done for the
     first index at position `index[:, 0]`.
 
@@ -111,6 +144,7 @@ def add_self_loops_to_edge_indices(edge_indices, *args, remove_duplicates: bool
         args (np.ndarray): Edge related value arrays to be changed accordingly of shape `(N, ...)`.
         remove_duplicates (bool): Remove duplicate edge indices. Default is True.
         sort_indices (bool): Sort final edge indices. Default is True.
+        fill_value (int): Value to initialize edge values with.
 
     Returns:
         edge_indices: Sorted index list with self-loops. Optionally (edge_indices, edge_values) if edge_values are not
@@ -118,18 +152,18 @@ def add_self_loops_to_edge_indices(edge_indices, *args, remove_duplicates: bool
     """
     clean_edge = [x for x in args]
     max_ind = np.max(edge_indices)
-    self_loops = np.arange(max_ind + 1, dtype=np.int)
+    self_loops = np.arange(max_ind + 1, dtype="int")
     self_loops = np.concatenate([np.expand_dims(self_loops, axis=-1), np.expand_dims(self_loops, axis=-1)], axis=-1)
     added_loops = np.concatenate([edge_indices, self_loops], axis=0)
     clean_index = added_loops
     for i, x in enumerate(clean_edge):
         edge_loops_shape = [self_loops.shape[0]] + list(x.shape[1:]) if len(x.shape) > 1 else [
             self_loops.shape[0]]
-        edge_loops = np.ones(edge_loops_shape)
+        edge_loops = np.full(edge_loops_shape, fill_value=fill_value, dtype=x.dtype)
         clean_edge[i] = np.concatenate([x, edge_loops], axis=0)
     if remove_duplicates:
         un, unis = np.unique(clean_index, return_index=True, axis=0)
-        mask_all = np.zeros(clean_index.shape[0], dtype=np.bool)
+        mask_all = np.zeros(clean_index.shape[0], dtype="bool")
         mask_all[unis] = True
         mask_all[:edge_indices.shape[0]] = True  # keep old indices untouched
         # clean_index = clean_index[unis]
@@ -176,7 +210,7 @@ def add_edges_reverse_indices(edge_indices, *args, remove_duplicates: bool = Tru
 
     if remove_duplicates:
         un, unis = np.unique(clean_index, return_index=True, axis=0)
-        mask_all = np.zeros(clean_index.shape[0], dtype=np.bool)
+        mask_all = np.zeros(clean_index.shape[0], dtype="bool")
         mask_all[unis] = True
         mask_all[:edge_indices.shape[0]] = True  # keep old indices untouched
         clean_index = clean_index[mask_all]
@@ -225,72 +259,90 @@ def sort_edge_indices(edge_indices, *args):
         return ind2
 
 
-def make_adjacency_from_edge_indices(edge_indices, edge_values=None):
+def make_adjacency_from_edge_indices(edge_indices, edge_values=None, shape=None):
     r"""Make adjacency as sparse matrix from a list or ``np.ndarray`` of edge_indices and possible values.
     Not for batches, only for single instance.
 
     Args:
         edge_indices (np.ndarray): List of edge indices of shape `(N, 2)`
         edge_values (np.ndarray): List of possible edge values of shape `(N, )`
+        shape (tuple): Shape of the sparse matrix. Default is None.
 
     Returns:
         scipy.coo.coo_matrix: Sparse adjacency matrix.
     """
-    if edge_values is None:
-        edge_values = np.ones(edge_indices.shape[0])
-    # index_min = np.min(edge_indices)
-    index_max = np.max(edge_indices)
     row = np.array(edge_indices[:, 0])
     col = np.array(edge_indices[:, 1])
+    if edge_values is None:
+        edge_values = np.ones(edge_indices.shape[0])
+    if shape is None:
+        edi_max = np.max(edge_indices)
+        shape = (edi_max + 1, edi_max + 1)
     data = edge_values
-    out_adj = sp.coo_matrix((data, (row, col)), shape=(index_max + 1, index_max + 1))
+    out_adj = sp.coo_matrix((data, (row, col)), shape=shape)
     return out_adj
 
 
-def get_angle_indices(idx, check_sorted: bool = True):
-    r"""Compute index list for edge-pairs forming an angle. Requires sorted indices.
-    Not for batches, only for single instance.
+def get_angle_indices(idx, check_sorted: bool = True, allow_multi_edges: bool = False):
+    r"""Compute index list for edge-pairs forming an angle. Not for batches, only for single instance.
 
     Args:
         idx (np.ndarray): List of edge indices referring to nodes of shape `(N, 2)`
-        check_sorted (bool): Whether to check inf indices are sorted. Default is True.
+        check_sorted (bool): Whether to to sort for new angle indices. Default is True.
+        allow_multi_edges (bool): Whether to keep angle pairs with :math:`i=k` in the tuple :math:`(i, j, k)`
 
     Returns:
         tuple: idx, idx_ijk, idx_ijk_ij
 
         - idx (np.ndarray): Edge indices referring to nodes of shape `(N, 2)`.
         - idx_ijk (np.ndarray): Indices of nodes forming an angle as i<-j<-k of shape `(M, 3)`.
-        - idx_ijk_ij (np.ndarray): Indices for an angle referring to edges of shape `(M, 2)`.
+        - idx_ijk_ij (np.ndarray): Indices for angle pairs referring to edges of shape `(M, 2)`.
     """
-    # Verify sorted
-    if check_sorted:
-        order1 = np.argsort(idx[:, 1], axis=0, kind='mergesort')  # stable!
-        ind1 = idx[order1]
-        order2 = np.argsort(ind1[:, 0], axis=0, kind='mergesort')
-        ind2 = ind1[order2]
-        if not np.array_equal(idx, ind2):
-            raise ValueError("ERROR:kgcnn: Indices need to be sorted to compute angles from.")
-
-    pair_label = np.arange(len(idx))
+    if idx is None:
+        return None, None, None
+    if len(idx) == 0:
+        return np.array([]), np.array([]), np.array([])
+    # Can swap here
     idx_i = idx[:, 0]
     idx_j = idx[:, 1]
-    uni_i, cnts_i = np.unique(idx_i, return_counts=True)
-    reps = cnts_i[idx_j]
-    idx_ijk_i = np.repeat(idx_i, reps)
-    idx_ijk_j = np.repeat(idx_j, reps)
-    idx_ijk_label_i = np.repeat(pair_label, reps)
-    idx_j_tagged = np.concatenate([np.expand_dims(idx_j, axis=-1), np.expand_dims(pair_label, axis=-1)], axis=-1)
-    idx_ijk_k_tagged = np.concatenate([idx_j_tagged[idx_i == x] for x in idx_j])  # This is not fully vectorized
-    idx_ijk_k = idx_ijk_k_tagged[:, 0]
-    idx_ijk_label_j = idx_ijk_k_tagged[:, 1]
-    back_and_forth = idx_ijk_i != idx_ijk_k
-    idx_ijk = np.concatenate([np.expand_dims(idx_ijk_i, axis=-1),
-                              np.expand_dims(idx_ijk_j, axis=-1),
-                              np.expand_dims(idx_ijk_k, axis=-1)], axis=-1)
-    idx_ijk = idx_ijk[back_and_forth]
-    idx_ijk_ij = np.concatenate([np.expand_dims(idx_ijk_label_i, axis=-1),
-                                 np.expand_dims(idx_ijk_label_j, axis=-1)], axis=-1)
-    idx_ijk_ij = idx_ijk_ij[back_and_forth]
+    label_ij = np.arange(len(idx))
+    # Direction of edges is i<-j here, so angles around i<-j<-k
+    idx_ijk = []  # index triples that form an angle as i<-j<-k
+    idx_ijk_ij = []  # New indices that refer to edges to form an angle as i',j'   i':(i, j) j':(j, k)
+    max_j = np.amax(idx_j)
+    for n in range(max_j + 1):
+        mask_jn = idx_j == n
+        mask_in = idx_i == n
+        pos_i = idx_i[mask_jn]
+        pos_j = idx_j[mask_jn]
+        pos_k = idx_j[mask_in]
+        label_i = label_ij[mask_jn]
+        label_k = label_ij[mask_in]
+        if len(pos_i) == 0 or len(pos_j) == 0 or len(pos_k) == 0:
+            continue
+        # All combos around j
+        combos_ik = np.concatenate(
+            [np.expand_dims(np.repeat(np.expand_dims(pos_i, axis=1), len(pos_k), axis=1), axis=-1),
+             np.expand_dims(np.repeat(np.expand_dims(pos_j, axis=1), len(pos_k), axis=1), axis=-1),
+             np.expand_dims(np.repeat(np.expand_dims(pos_k, axis=0), len(pos_i), axis=0), axis=-1)], axis=-1)
+        combos_label = np.concatenate(
+            [np.expand_dims(np.repeat(np.expand_dims(label_i, axis=1), len(label_k), axis=1), axis=-1),
+             np.expand_dims(np.repeat(np.expand_dims(label_k, axis=0), len(label_i), axis=0), axis=-1)], axis=-1)
+
+        mask_combos = np.ones(combos_label.shape[:2], dtype="bool")
+        if not allow_multi_edges:
+            mask_combos = np.array(combos_ik[:, :, 0] != combos_ik[:, :, -1], dtype="bool")
+        idx_ijk.append(combos_ik[mask_combos])
+        idx_ijk_ij.append(combos_label[mask_combos])
+    idx_ijk = np.concatenate(idx_ijk, axis=0)
+    idx_ijk_ij = np.concatenate(idx_ijk_ij, axis=0)
+    if check_sorted:
+        order1 = np.argsort(idx_ijk_ij[:, 1], axis=0, kind='mergesort')  # stable!
+        idx_ijk_ij = idx_ijk_ij[order1]
+        idx_ijk = idx_ijk[order1]
+        order2 = np.argsort(idx_ijk_ij[:, 0], axis=0, kind='mergesort')
+        idx_ijk = idx_ijk[order2]
+        idx_ijk_ij = idx_ijk_ij[order2]
 
     return idx, idx_ijk, idx_ijk_ij
 
@@ -308,6 +360,10 @@ def get_angle(coord, indices):
     Returns:
         np.ndarray: List of angles matching indices `(M, 1)`.
     """
+    if coord is None or indices is None:
+        return
+    if len(indices) == 0:
+        return np.array([])
     xi = coord[indices[:, 0]]
     xj = coord[indices[:, 1]]
     xk = coord[indices[:, 2]]
@@ -493,3 +549,30 @@ def define_adjacency_from_distance(distance_matrix, max_distance=np.inf, max_nei
 
     graph_indices = graph_indices[graph_adjacency]
     return graph_adjacency, graph_indices
+
+
+def compute_reverse_edges_index_map(edge_idx):
+    r"""Computes the index map of the reverse edge for each of the edges if available. This can be used by a model
+    to directly select the corresponding edge of :math:`(j, i)` which is :math:`(i, j)`.
+    Edges that do not have a reverse pair get a `nan` as map index.
+    If there are multiple edges, the first encounter is assigned.
+
+    Args:
+        edge_idx (np.ndarray): Array of edge indices of shape `(N, 2)`.
+
+    Returns:
+        np.ndarray: Map of reverse indices of shape `(N, )`.
+    """
+    if len(edge_idx) == 0:
+        return np.array([], dtype="int")
+
+    edge_idx_rev = np.flip(edge_idx, axis=-1)
+    edge_pos, rev_pos = np.where(
+        np.all(np.expand_dims(edge_idx, axis=1) == np.expand_dims(edge_idx_rev, axis=0), axis=-1))
+    # May have duplicates, find unique.
+    ege_pos_uni, uni_pos = np.unique(edge_pos, return_index=True)
+    rev_pos_uni = rev_pos[uni_pos]
+    edge_map = np.empty(len(edge_idx), dtype="int")
+    edge_map.fill(np.nan)
+    edge_map[ege_pos_uni] = rev_pos_uni
+    return edge_map
