@@ -7,8 +7,10 @@ from scipy import special as sp
 
 import tensorflow as tf
 from kgcnn.utils.adj import get_angle_indices
-from kgcnn.layers.geom import SphericalBasisLayer, NodeDistance, EdgeAngle
+from kgcnn.layers.geom import NodeDistanceEuclidean, EdgeAngle, NodePosition, EdgeAngle2
+from kgcnn.layers.conv.dimenet_conv import SphericalBasisLayer
 from kgcnn.layers.geom import BesselBasisLayer
+from kgcnn.layers.modules import LazySubtract
 
 
 class TestSphericalBasisLayer(unittest.TestCase):
@@ -272,9 +274,10 @@ class TestSphericalBasisLayer(unittest.TestCase):
 
                 return cbf*rbf_env
 
-
-        dist = NodeDistance()([rag_x, rag_ei])
-        angs = EdgeAngle()([rag_x, rag_ei, rag_a])
+        a, b = NodePosition()([rag_x, rag_ei])
+        dist = NodeDistanceEuclidean()([a, b])
+        vec = LazySubtract()([a, b])
+        angs = EdgeAngle()([vec, rag_a])
         bessel = SphericalBasisLayer(10,10,5.0)([dist, angs, rag_a])
 
         test = np.array(SphericalBasisLayerOrginal(10,10,5)([dist[0,:,0], angs[0,:,0], rag_a[0,:,1]]))
@@ -288,7 +291,7 @@ class TestSphericalBasisLayer(unittest.TestCase):
         # print(np.max(valid1))
 
         # For small argument and large order there are some rof errors
-        # So we only have < 0.05 here
+        # So we only have < 0.1 here
         self.assertTrue(np.max(valid) < 0.05)
         self.assertTrue(np.max(valid1) < 0.05)
 
@@ -332,7 +335,8 @@ class TestBesselBasisLayer(unittest.TestCase):
 
         rag_x = tf.RaggedTensor.from_row_lengths(np.concatenate([x,x1]), np.array([len(x),len(x1)]))
         rag_ei = tf.RaggedTensor.from_row_lengths(np.concatenate([ei,ei1]), np.array([len(ei),len(ei1)]))
-        dist = NodeDistance()([rag_x, rag_ei])
+        a, b = NodePosition()([rag_x, rag_ei])
+        dist = NodeDistanceEuclidean()([a, b])
         bessel = BesselBasisLayer(10,5.0)(dist)
 
         class Envelope(tf.keras.layers.Layer):
